#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.db.extensions.auth.model`` -- Schema Definition
"""

from sqlalchemy import *
from sqlalchemy.orm import relationship
from sqlalchemy.ext.associationproxy import association_proxy

import edbob
from edbob.db.model import Base, uuid_column
from edbob.db.extensions.contact import Person
from edbob.sqlalchemy import getset_factory


__all__ = ['Role', 'User', 'UserRole', 'Permission']


class Permission(Base):
    """
    Represents the fact that a particular :class:`Role` is allowed to do a
    particular type of thing.
    """

    __tablename__ = 'permissions'

    role_uuid = Column(String(32), ForeignKey('roles.uuid'), primary_key=True)
    permission = Column(String(50), primary_key=True)

    def __repr__(self):
        return "<Permission: %s, %s>" % (self.role, self.permission)

    def __str__(self):
        return str(self.permission or '')


class UserRole(Base):
    """
    Represents the association between a :class:`User` and a :class:`Role`.
    """

    __tablename__ = 'users_roles'

    uuid = uuid_column()
    user_uuid = Column(String(32), ForeignKey('users.uuid'))
    role_uuid = Column(String(32), ForeignKey('roles.uuid'))

    def __repr__(self):
        return "<UserRole: %s : %s>" % (self.user, self.role)


class Role(Base):
    """
    Represents a role within the system; used to manage permissions.
    """

    __tablename__ = 'roles'

    uuid = uuid_column()
    name = Column(String(25), nullable=False, unique=True)

    _permissions = relationship(
        Permission, backref='role',
        cascade='save-update, merge, delete, delete-orphan')
    permissions = association_proxy('_permissions', 'permission',
                                    creator=lambda x: Permission(permission=x),
                                    getset_factory=getset_factory)

    _users = relationship(
        UserRole, backref='role',
        cascade='save-update, merge, delete, delete-orphan')
    users = association_proxy('_users', 'user',
                              creator=lambda x: UserRole(user=x),
                              getset_factory=getset_factory)

    def __repr__(self):
        return "<Role: %s>" % self.name

    def __str__(self):
        return str(self.name or '')


class User(Base):
    """
    Represents a user of the system.  This may or may not correspond to a real
    person, i.e. some users may exist solely for automated tasks.
    """

    __tablename__ = 'users'

    uuid = uuid_column()
    username = Column(String(25), nullable=False, unique=True)
    password = Column(String(60))
    salt = Column(String(29))
    person_uuid = Column(String(32), ForeignKey('people.uuid'))

    _roles = relationship(UserRole, backref='user')
    roles = association_proxy(
        '_roles', 'role',
        creator=lambda x: UserRole(role=x),
        getset_factory=getset_factory)

    def __repr__(self):
        return "<User: %s>" % self.username

    def __str__(self):
        return str(self.username or '')

    @property
    def display_name(self):
        """
        Returns the user's ``person.display_name``, if present, otherwise the
        ``username``.
        """
        if self.person and self.person.display_name:
            return self.person.display_name
        return self.username


User.person = relationship(
    Person,
    back_populates='user',
    uselist=False)

Person.user = relationship(
    User,
    back_populates='person',
    uselist=False)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.forms.formalchemy.fields`` -- Field Classes
"""

import formalchemy


__all__ = ['AssociationProxyField']


def AssociationProxyField(name, **kwargs):
    """
    Returns a :class:`Field` class which is aware of SQLAlchemy association
    proxies.
    """

    class ProxyField(formalchemy.Field):

        def sync(self):
            if not self.is_readonly():
                setattr(self.parent.model, self.name,
                        self.renderer.deserialize())

    kwargs.setdefault('value', lambda x: getattr(x, name))
    return ProxyField(name, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.forms.formalchemy.renderers`` -- Field Renderers
"""

import formalchemy

from pyramid.renderers import render


__all__ = ['AutocompleteFieldRenderer', 'EnumFieldRenderer',
           'YesNoFieldRenderer']


def AutocompleteFieldRenderer(service_url, field_value=None, field_display=None, width='300px'):
    """
    Autocomplete renderer.
    """

    class AutocompleteFieldRenderer(formalchemy.fields.FieldRenderer):

        @property
        def focus_name(self):
            return self.name + '-textbox'

        @property
        def needs_focus(self):
            return not bool(self.value or field_value)

        def render(self, **kwargs):
            kwargs.setdefault('field_name', self.name)
            kwargs.setdefault('field_value', self.value or field_value)
            kwargs.setdefault('field_display', self.raw_value or field_display)
            kwargs.setdefault('service_url', service_url)
            kwargs.setdefault('width', width)
            return render('/forms/field_autocomplete.mako', kwargs)

    return AutocompleteFieldRenderer


def EnumFieldRenderer(enum):
    """
    Adds support for enumeration fields.
    """

    class Renderer(formalchemy.fields.SelectFieldRenderer):
        
        def render_readonly(self, **kwargs):
            value = self.raw_value
            if value is None:
                return ''
            if value in enum:
                return enum[value]
            return str(value)

        def render(self, **kwargs):
            opts = [(enum[x], x) for x in sorted(enum)]
            return formalchemy.fields.SelectFieldRenderer.render(self, opts, **kwargs)

    return Renderer


class YesNoFieldRenderer(formalchemy.fields.CheckBoxFieldRenderer):

    def render_readonly(self, **kwargs):
        value = self.raw_value
        if value is None:
            return ''
        return 'Yes' if value else 'No'



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.win32`` -- Stuff for Microsoft Windows
"""

import sys
import subprocess
import logging

if sys.platform == 'win32': # docs should build for everyone
    import pywintypes
    import win32api
    import win32con
    import win32event
    import win32file
    import win32print
    import win32service
    import win32serviceutil
    import winerror

import edbob


log = logging.getLogger(__name__)


class Service(win32serviceutil.ServiceFramework):
    """
    Base class for Windows service implementations.
    """

    appname = 'edbob'

    def __init__(self, args):
        """
        Constructor.
        """

        win32serviceutil.ServiceFramework.__init__(self, args)

        # Create "wait stop" event, for main worker loop.
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)

    def Initialize(self):
        """
        Service initialization.
        """

        # Read configuration file(s).
        edbob.init(self.appname, service=self._svc_name_)

        return True

    def SvcDoRun(self):
        """
        This method is invoked when the service starts.
        """

        import servicemanager

        # Write start occurrence to Windows Event Log.
        servicemanager.LogMsg(
            servicemanager.EVENTLOG_INFORMATION_TYPE,
            servicemanager.PYS_SERVICE_STARTED,
            (self._svc_name_, ''))

        # Figure out what we're supposed to be doing.
        if self.Initialize():

            # Wait infinitely for stop request, while threads do their thing.
            log.info("SvcDoRun: All threads started; waiting for stop request.")
            win32event.WaitForSingleObject(self.hWaitStop, win32event.INFINITE)
            log.info("SvcDoRun: Stop request received.")

        else: # Nothing to be done...
            msg = "Nothing to do!  (Initialization failed.)"
            servicemanager.LogWarningMsg(msg)
            log.warning("SvcDoRun: %s" % msg)
            self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)

        # Write stop occurrence to Windows Event Log.
        servicemanager.LogMsg(
            servicemanager.EVENTLOG_INFORMATION_TYPE,
            servicemanager.PYS_SERVICE_STOPPED,
            (self._svc_name_, ''))

    def SvcStop(self):
        """
        This method is invoked when the service is requested to stop itself.
        """

        # Let the SCM know we're trying to stop.
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)

        # Let worker loop know its job is done.
        win32event.SetEvent(self.hWaitStop)


def RegDeleteTree(key, subkey):
    """
    This is a clone of ``win32api.RegDeleteTree()``, since that apparently
    requires Vista or later.
    """

    def delete_contents(key):
        subkeys = []
        for name, reserved, class_, mtime in win32api.RegEnumKeyEx(key):
            subkeys.append(name)
        for subkey_name in subkeys:
            subkey = win32api.RegOpenKeyEx(key, subkey_name, 0, win32con.KEY_ALL_ACCESS)
            delete_contents(subkey)
            win32api.RegCloseKey(subkey)
            win32api.RegDeleteKey(key, subkey_name)
        values = []
        i = 0
        while True:
            try:
                name, value, type_ = win32api.RegEnumValue(key, i)
            except pywintypes.error, e:
                if e[0] == winerror.ERROR_NO_MORE_ITEMS:
                    break
            values.append(name)
            i += 1
        for value in values:
            win32api.RegDeleteValue(key, value)

    orig_key = key
    try:
        key = win32api.RegOpenKeyEx(orig_key, subkey, 0, win32con.KEY_ALL_ACCESS)
    except pywintypes.error, e:
        if e[0] != winerror.ERROR_FILE_NOT_FOUND:
            raise
    else:
        delete_contents(key)
        win32api.RegCloseKey(key)
    try:
        win32api.RegDeleteKey(orig_key, subkey)
    except pywintypes.error, e:
        if e[0] == winerror.ERROR_FILE_NOT_FOUND:
            pass


def capture_output(command):
    """
    Runs ``command`` and returns any output it produces.
    """

    # We *need* to pipe ``stdout`` because that's how we capture the output of
    # the ``hg`` command.  However, we must pipe *all* handles in order to
    # prevent issues when running as a GUI but *from* the Windows console.
    # See also: http://bugs.python.org/issue3905
    kwargs = dict(stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    output = subprocess.Popen(command, **kwargs).communicate()[0]
    return output


def delayed_auto_start_service(name):
    """
    Configures the Windows service named ``name`` such that its startup type is
    "Automatic (Delayed Start)".

    .. note::
       It is assumed that the service is already configured to start
       automatically.  This function only modifies the service so that its
       automatic startup is delayed.
    """

    hSCM = win32service.OpenSCManager(
        None,
        None,
        win32service.SC_MANAGER_ENUMERATE_SERVICE)

    hService = win32service.OpenService(
        hSCM,
        name,
        win32service.SERVICE_CHANGE_CONFIG)

    win32service.ChangeServiceConfig2(
        hService,
        win32service.SERVICE_CONFIG_DELAYED_AUTO_START_INFO,
        True)

    win32service.CloseServiceHandle(hService)
    win32service.CloseServiceHandle(hSCM)


def execute_service_command(module, command, *args):
    """
    Executes ``command`` against the Windows service contained in ``module``.

    ``module`` must be a proper module object, which is assumed to implement a
    command line interface when invoked directly by the Python interpreter, a
    la ``win32serviceutil.HandleCommandLine()``.

    ``command`` may be anything supported by ``HandleCommandLine()``, e.g.:

    * ``'install'``
    * ``'remove'``
    * ``'start'``
    * ``'stop'``
    * ``'restart'``

    ``args``, if present, are assumed to be "option" arguments and will precede
    ``command`` when the command line is constructed.
    """

    command = [command]
    if args:
        command = list(args) + command
    subprocess.call([sys.executable, module.__file__] + command)


def file_is_free(path):
    """
    Returns boolean indicating whether or not the file located at ``path`` is
    currently tied up in any way by another process.
    """

    # This code was borrowed from Nikita Nemkin:
    # http://stackoverflow.com/a/2848266

    handle = None
    try:
        handle = win32file.CreateFile(
            path,
            win32file.GENERIC_WRITE,
            0,
            None,
            win32file.OPEN_EXISTING,
            win32file.FILE_ATTRIBUTE_NORMAL,
            None)
        return True
    except pywintypes.error, e:
        if e[0] == winerror.ERROR_SHARING_VIOLATION:
            return False
        raise
    finally:
        if handle:
            win32file.CloseHandle(handle)


def send_data_to_printer(data, printer_name, job_name):
    """
    Create and submit a new print job (named ``job_name``) which sends ``data``
    directly to the Windows printer identified by ``printer_name``.  Returns
    the number of bytes actually written to the printer port.

    This is designed for sending command strings to Zebra label printers, but
    could potentially be useful for other situations as well.
    """

    printer = win32print.OpenPrinter(printer_name)
    assert printer
    assert win32print.StartDocPrinter(printer, 1, (job_name, None, None))
    win32print.StartPagePrinter(printer)
    num_bytes = win32print.WritePrinter(printer, data)
    win32print.EndPagePrinter(printer)
    win32print.EndDocPrinter(printer)
    win32print.ClosePrinter(printer)
    return num_bytes



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.db.extensions`` -- Database Extensions
"""

import logging
# from pkg_resources import iter_entry_points

import sqlalchemy.exc
from sqlalchemy import MetaData
# from sqlalchemy.orm import clear_mappers

# import migrate.versioning.api
# from migrate.versioning.schema import ControlledSchema

# import rattail
# from rattail.db import exc as exceptions
# from rattail.db import Session
# from rattail.db.classes import ActiveExtension
# from rattail.db.mappers import make_mappers
# from rattail.db.model import get_metadata
# from rattail.db.util import get_repository_path, get_repository_version

import edbob
import edbob.db
from edbob.db import exceptions
from edbob.db import Session
# from edbob.db.classes import ActiveExtension
from edbob.db.model import Base, ActiveExtension
# from edbob.db.util import (
#     get_database_version,
#     get_repository_path,
#     get_repository_version,
#     )
from edbob.modules import import_module_path
from edbob.util import entry_point_map, requires_impl


log = logging.getLogger(__name__)


class Extension(edbob.Object):
    """
    Base class for schema/ORM extensions.
    """

    # Set this to a list of strings (extension names) as needed within your
    # derived class.
    required_extensions = []

    # You can set this to any dotted module path you like.  If unset a default
    # will be assumed, of the form ``<path.to.extension>.model`` (see
    # :meth:`Extension.get_models_module()` for more info).
    model_module = ''

    # You can set this to any dotted module path you like.  If unset a default
    # will be assumed, of the form ``<path.to.extension>.enum`` (see
    # :meth:`Extension.get_enum_module()` for more info).
    enum_module = ''

    # @property
    # @requires_impl(is_property=True)
    # def name(self):
    #     """
    #     The name of the extension.
    #     """
    #     pass

    # @property
    # @requires_impl(is_property=True)
    # def schema(self):
    #     """
    #     Should return a reference to the extension's ``schema`` module, which
    #     is assumed to be a SQLAlchemy-Migrate repository.
    #     """
    #     pass

    # def add_class(self, cls):
    #     """
    #     Convenience method for use in :meth:`extend_classes()`.
    #     """

    #     from edbob.db import classes

    #     name = cls.__name__
    #     edbob.graft(classes, {name:cls}, name)

    # def extend_classes(self):
    #     """
    #     Any extra classes provided by the extension should be added to the ORM
    #     whenever this method is called.

    #     Note that the :meth:`add_class()` convenience method is designed to be
    #     used when adding classes.
    #     """
    #     pass

    def extend_framework(self):
        """
        Extends the framework...
        """

        edbob.graft(edbob, self.get_model_module())
        enum = self.get_enum_module()
        if enum:
            edbob.graft(edbob, enum)

    # def extend_mappers(self, metadata):
    #     """
    #     All SQLAlchemy mapping to be done by the extension should be done
    #     within this method.

    #     Any extra classes the extension provides will typically be mapped here.
    #     Any manipulation the extension needs to perform on the ``edbob`` core
    #     ORM should be done here as well.
    #     """
    #     pass

    def get_metadata(self, recurse=False):
        """
        Returns a :class:`sqlalchemy.MetaData` instance containing the schema
        definition for the extension.

        If ``recurse`` evaluates to true, then tables from any extensions upon
        which this one relies will be included as well.
        """

        meta = MetaData()
        self.populate_metadata(meta, recurse)
        return meta

    def get_enum_module(self):
        """
        Imports and returns a reference to the Python module providing
        enumeration data for the extension (if one exists).

        :attr:`Extension.enum_module` is first consulted to determine the
        dotted module path.  If nothing is found there, a default path is
        constructed by appending ``'.enum'`` to the extension module's own
        dotted path.
        """

        if self.enum_module:
            module = self.enum_module
        else:
            module = str(self.__class__.__module__) + '.enum'
        try:
            return import_module_path(module)
        except ImportError:
            return None

    def get_model_module(self):
        """
        Imports and returns a reference to the Python module providing schema
        definition for the extension.

        :attr:`Extension.model_module` is first consulted to determine the
        dotted module path.  If nothing is found there, a default path is
        constructed by appending ``'.model'`` to the extension module's own
        dotted path.
        """

        if self.model_module:
            module = self.model_module
        else:
            module = str(self.__class__.__module__) + '.model'
        return import_module_path(module)

    def populate_metadata(self, metadata, recurse=False):
        """
        Populates ``metadata`` with tables provided by the extension.

        If ``recurse`` evaluates to true, then tables for any extension upon
        which this one relies will also be included.
        """

        if recurse:
            for name in self.required_extensions:
                ext = get_extension(name)
                ext.populate_metadata(metadata, True)

        model = self.get_model_module()
        for name in model.__all__:
            obj = getattr(model, name)
            if isinstance(obj, type) and issubclass(obj, model.Base):
                if obj.__tablename__ not in metadata.tables:
                    obj.__table__.tometadata(metadata)

    # def remove_class(self, name):
    #     """
    #     Convenience method for use in :meth:`restore_classes()`.
    #     """

    #     from edbob.db import classes

    #     if name in classes.__all__:
    #         classes.__all__.remove(name)
    #     if hasattr(classes, name):
    #         del classes.__dict__[name]

    # def restore_classes(self):
    #     """
    #     This method should remove any extra classes which were added within
    #     :meth:`extend_classes()`.  Note that there is a :meth:`remove_class()`
    #     method for convenience in doing so.
    #     """
    #     pass


def activate_extension(extension, engine=None):
    """
    Activates the :class:`Extension` instance represented by ``extension``
    (which can be the actual instance, or the extension's name) by installing
    its schema and registering it within the database, and immediately applies
    it to the current ORM/API.

    If ``engine`` is not provided, then :attr:`edbob.db.engine` is assumed.
    """

    if engine is None:
        engine = edbob.db.engine

    if not isinstance(extension, Extension):
        extension = get_extension(extension)

    # Skip all this if already active.
    if extension_active(extension, engine):
        return

    log.debug("Activating extension: %s" % extension.name)

    # Activate all required extensions first.
    for name in extension.required_extensions:
        activate_extension(name, engine)

    # Install schema for this extension.
    install_extension_schema(extension, engine)

    # Add ActiveExtension record for this extension.
    session = Session(bind=engine)
    if not session.query(ActiveExtension).get(extension.name):
        session.add(ActiveExtension(name=extension.name))
        session.commit()
    session.close()

    # merge_extension_metadata(extension)
    # extension.extend_classes()
    # extension.extend_mappers(Base.metadata)
    extension.extend_framework()

    # Add extension to in-memory active extensions tracker.
    active_extensions(engine).append(extension.name)


_available_extensions = None
def available_extensions():
    """
    Returns the map of available :class:`Extension` classes, as determined by
    ``'edbob.db.extensions'`` entry points..
    """

    global _available_extensions

    if _available_extensions is None:
        _available_extensions = entry_point_map('edbob.db.extensions')
    return _available_extensions


def deactivate_extension(extension, engine=None):
    """
    Uninstalls an extension's schema from a database.

    If ``engine`` is not provided, :attr:`edbob.db.engine` is assumed.
    """

    if engine is None:
        engine = edbob.db.engine

    if not isinstance(extension, Extension):
        extension = get_extension(extension)

    log.debug("Deactivating extension: %s" % extension.name)
    active = active_extensions(engine)
    if extension.name in active:
        active.remove(extension.name)

    session = Session(bind=engine)
    ext = session.query(ActiveExtension).get(extension.name)
    if ext:
        session.delete(ext)
        session.commit()
    session.close()

    uninstall_extension_schema(extension, engine)
    # unmerge_extension_metadata(extension)
    # extension.restore_classes()

    # clear_mappers()
    # make_mappers(rattail.metadata)
    # for name in sorted(_active_extensions, extension_sorter(_active_extensions)):
    #     _active_extensions[name].extend_mappers(rattail.metadata)


def extend_framework():
    """
    Attempts to connect to the primary database and, if successful, inspects it
    to determine which extensions are active within it.  Any such extensions
    found will be used to extend the ORM/API in-place.
    """

    # Do we even have an engine?
    engine = edbob.db.engine
    if not engine:
        return

    # Check primary database connection.
    try:
        engine.connect()
    except sqlalchemy.exc.OperationalError:
        return

    # # Check database version to see if core schema is installed.
    # try:
    #     db_version = get_database_version(engine)
    # except exceptions.CoreSchemaNotInstalled:
    #     return

    # Since extensions may depend on one another, we must first retrieve the
    # list of active extensions' names from the database and *then* sort them
    # according to their stated dependencies.  (This information is only known
    # after instantiating the extensions.)

    session = Session()
    try:
        active = session.query(ActiveExtension).all()
    except sqlalchemy.exc.ProgrammingError:
        session.close()
        return

    extensions = {}
    for ext in active:
        extensions[ext.name] = get_extension(ext.name)
    session.close()

    for name in sorted(extensions, extension_sorter(extensions)):
        log.info("Applying active extension: %s" % name)
        ext = extensions[name]
        # merge_extension_metadata(ext)
        # ext.extend_classes()
        # ext.extend_mappers(rattail.metadata)
        ext.extend_framework()
        _active_extensions[name] = ext


def extension_active(extension, engine=None):
    """
    Returns boolean indicating whether or not the given ``extension`` is active
    within a database.

    If ``engine`` is not provided, :attr:`edbob.db.engine` is assumed.
    """

    if not engine:
        engine = edbob.db.engine

    if not isinstance(extension, Extension):
        extension = get_extension(extension)
    return extension.name in active_extensions(engine)


def extension_sorter(extensions):
    """
    Returns a function to be used for sorting extensions according to their
    inter-dependencies.  ``extensions`` should be a dictionary containing the
    extensions which are to be sorted.
    """

    def sorter(name_x, name_y):
        ext_x = extensions[name_x]
        ext_y = extensions[name_y]

        if name_y in ext_x.required_extensions:
            return 1
        if name_x in ext_y.required_extensions:
            return -1

        if ext_x.required_extensions and not ext_y.required_extensions:
            return 1
        if ext_y.required_extensions and not ext_x.required_extensions:
            return -1
        return 0

    return sorter        


def get_extension(name):
    """
    Returns a :class:`Extension` instance, according to ``name``.  An error is
    raised if the extension cannot be found.
    """

    extensions = available_extensions()
    if name in extensions:
        return extensions[name]()
    raise exceptions.ExtensionNotFound(name)


def install_extension_schema(extension, engine=None):
    """
    Installs an extension's schema to the database and adds version control for
    it.  ``extension`` must be a valid :class:`Extension` instance.

    If ``engine`` is not provided, :attr:`edbob.db.engine` is assumed.
    """

    if engine is None:
        engine = edbob.db.engine

    # # Extensions aren't required to provide metadata...
    # ext_meta = extension.get_metadata()
    # if not ext_meta:
    #     return

    # # ...but if they do they must also provide a SQLAlchemy-Migrate repository.
    # assert extension.schema, "Extension does not implement 'schema': %s" % extension.name
        
    # meta = edbob.db.metadata
    # for table in meta.sorted_tables:
    #     table.tometadata(ext_meta)
    # for table in ext_meta.sorted_tables:
    #     if table.name not in meta.tables:
    #         table.create(bind=engine, checkfirst=True)

    # TODO: This sucks, please fix.
    # edbob.db.Base.metadata.create_all(engine)

    # meta = MetaData(engine)
    # for tables in (edbob.db.iter_tables(), extension.iter_tables()):
    #     for table in tables:
    #         table.tometadata(meta)
    # meta.create_all()

    core_meta = edbob.db.get_core_metadata()
    ext_meta = extension.get_metadata(recurse=True)
    for table in ext_meta.sorted_tables:
        table.tometadata(core_meta)
    core_meta.create_all(engine)

    # migrate.versioning.api.version_control(
    #     str(engine.url), get_repository_path(extension), get_repository_version(extension))


def merge_extension_metadata(ext):
    """
    Merges an extension's metadata with the global ``edbob.db.metadata``
    instance.

    .. note::
       ``edbob`` uses this internally; you should not need to.
    """

    ext_meta = ext.get_metadata()
    if not ext_meta:
        return
    meta = Base.metadata
    for table in meta.sorted_tables:
        table.tometadata(ext_meta)
    for table in ext_meta.sorted_tables:
        if table.name not in meta.tables:
            table.tometadata(meta)


def uninstall_extension_schema(extension, engine=None):
    """
    Uninstalls an extension's tables from the database represented by
    ``engine`` (or :attr:`edbob.db.engine` if none is provided), and removes
    SQLAlchemy-Migrate version control for the extension.
    """

    if engine is None:
        engine = edbob.db.engine
    
    # ext_meta = extension.get_metadata()
    # if not ext_meta:
    #     return
    
    # schema = ControlledSchema(engine, get_repository_path(extension))
    # engine.execute(schema.table.delete().where(
    #         schema.table.c.repository_id == schema.repository.id))

    # meta = get_metadata()
    # for table in meta.sorted_tables:
    #     table.tometadata(ext_meta)
    # for table in reversed(ext_meta.sorted_tables):
    #     if table.name not in meta.tables:
    #         table.drop(bind=engine)

    # core_meta = edbob.db.get_core_metadata()
    # ext_meta = extension.get_metadata()
    # for table in ext_meta.sorted_tables:
    #     table.tometadata(core_meta)
    # core_meta.create_all(engine)

    # core_meta = edbob.db.get_core_metadata()
    # ext_meta = extension.get_metadata()
    # for table in ext_meta.sorted_tables:
    #     table.tometadata(core_meta)
    # for table in reversed(core_meta.sorted_tables):
    #     if table in ext_meta:
    #         table.drop(engine)

    core_meta = edbob.db.get_core_metadata()
    ext_fullmeta = extension.get_metadata(True)
    for table in ext_fullmeta.sorted_tables:
        table.tometadata(core_meta)
    ext_meta = extension.get_metadata()
    for table in reversed(core_meta.sorted_tables):
        if table in ext_meta:
            table.drop(engine)


# def unmerge_extension_metadata(extension):
#     """
#     Removes an extension's metadata from the global ``rattail.metadata``
#     instance.
#     """

#     ext_meta = extension.get_metadata()
#     if not ext_meta:
#         return

#     meta = rattail.metadata
#     ext_tables = ext_meta.tables.keys()
#     for table in reversed(meta.sorted_tables):
#         if table.name in ext_tables:
#             meta.remove(table)


# # def merge_extension_permissions(extension):
# #     '''
# #     Helper function to merge an extension's permission definitions with those of
# #     the framework.  (This should only be called by the framework itself.)
# #     '''
# #     from rattail.v1.perms import permissions
# #     log.debug('Merging permissions from extension: %s' % extension.name)
# #     for group_name in extension.permissions:
# #         if group_name not in permissions:
# #             permissions[group_name] = extension.permissions[group_name]
# #         elif extension.permissions[group_name][0] != permissions[group_name][0]:
# #             log.warning("Extension '%s' tries to override UUID of permission group '%s' (but is denied)" % (
# #                     extension.name, group_name))
# #         else:
# #             # Extensions may override permission group display names.
# #             if extension.permissions[group_name][1]:
# #                 permissions[group_name][1] = extension.permissions[group_name][1]
# #             perms = permissions[group_name][2]
# #             ext_perms = extension.permissions[group_name][2]
# #             for perm_name in ext_perms:
# #                 if perm_name not in perms:
# #                     perms[perm_name] = ext_perms[perm_name]
# #                 elif ext_perms[perm_name][0] != perms[perm_name][0]:
# #                     log.warning("Extension '%s' tries to override UUID of permission '%s' (but is denied)" % (
# #                             extension.name, '.'.join((group_name, perm_name))))
# #                 else:
# #                     # Extensions may override permission display names.
# #                     if ext_perms[perm_name][1]:
# #                         perms[perm_name][1] = ext_perms[perm_name][1]


_active_extensions = {}
def active_extensions(engine=None):
    """
    Returns a list of names for extensions which are active within a database.

    If ``engine`` is not provided, ``edbob.db.engine`` is assumed.
    """

    if not engine:
        engine = edbob.db.engine

    exts = _active_extensions.get(engine.url)
    if exts:
        return exts

    session = Session()
    q = session.query(ActiveExtension.name)
    exts = [x[0] for x in q]
    session.close()

    _active_extensions[engine.url] = exts
    return exts



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.console`` -- Console-Specific Stuff
"""

import sys
import progressbar

import edbob


class Progress(edbob.Object):
    """
    Provides a console-based progress bar.
    """

    def __init__(self, message, maximum):
        sys.stderr.write("\n%s...(%u total)\n" % (message, maximum))
        widgets = [progressbar.Percentage(), ' ', progressbar.Bar(), ' ', progressbar.ETA()]
        self.progress = progressbar.ProgressBar(maxval=maximum, widgets=widgets).start()

    def update(self, value):
        self.progress.update(value)
        return True

    def destroy(self):
        sys.stderr.write("\n")



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.users`` -- User Views
"""

from webhelpers.html import tags
from webhelpers.html.builder import HTML

import formalchemy
from formalchemy.fields import SelectFieldRenderer

import edbob
from edbob.db import auth
from edbob.pyramid import Session
from edbob.pyramid.views import SearchableAlchemyGridView, CrudView


class UsersGrid(SearchableAlchemyGridView):

    mapped_class = edbob.User
    config_prefix = 'users'
    sort = 'username'

    def join_map(self):
        return {
            'person':
                lambda q: q.outerjoin(edbob.Person),
            }

    def filter_map(self):
        return self.make_filter_map(
            ilike=['username'],
            person=self.filter_ilike(edbob.Person.display_name))

    def filter_config(self):
        return self.make_filter_config(
            include_filter_username=True,
            filter_type_username='lk',
            include_filter_person=True,
            filter_type_person='lk')

    def sort_map(self):
        return self.make_sort_map(
            'username',
            person=self.sorter(edbob.Person.display_name))

    def grid(self):
        g = self.make_grid()
        g.configure(
            include=[
                g.username,
                g.person,
                ],
            readonly=True)
        if self.request.has_perm('users.read'):
            g.clickable = True
            g.click_route_name = 'user.read'
        if self.request.has_perm('users.update'):
            g.editable = True
            g.edit_route_name = 'user.update'
        if self.request.has_perm('users.delete'):
            g.deletable = True
            g.delete_route_name = 'user.delete'
        return g


def RolesFieldRenderer(request):

    class RolesFieldRenderer(SelectFieldRenderer):

        def render_readonly(self, **kwargs):
            roles = Session.query(edbob.Role)
            html = ''
            for uuid in self.value:
                role = roles.get(uuid)
                link = tags.link_to(
                    role.name, request.route_url('role.read', uuid=role.uuid))
                html += HTML.tag('li', c=link)
            html = HTML.tag('ul', c=html)
            return html

    return RolesFieldRenderer


class RolesField(formalchemy.Field):

    def __init__(self, name, **kwargs):
        kwargs.setdefault('value', self.get_value)
        kwargs.setdefault('options', self.get_options())
        kwargs.setdefault('multiple', True)
        super(RolesField, self).__init__(name, **kwargs)

    def get_value(self, user):
        return [x.uuid for x in user.roles]

    def get_options(self):
        q = Session.query(edbob.Role.name, edbob.Role.uuid)
        q = q.filter(edbob.Role.uuid != auth.guest_role(Session()).uuid)
        q = q.order_by(edbob.Role.name)
        return q.all()

    def sync(self):
        if not self.is_readonly():
            user = self.model
            roles = Session.query(edbob.Role)
            data = self.renderer.deserialize()
            user.roles = [roles.get(x) for x in data]
                

class _ProtectedPersonRenderer(formalchemy.FieldRenderer):

    def render_readonly(self, **kwargs):
        res = str(self.person)
        res += tags.hidden('User--person_uuid',
                           value=self.field.parent.person_uuid.value)
        return res


def ProtectedPersonRenderer(uuid):
    person = Session.query(edbob.Person).get(uuid)
    assert person
    return type('ProtectedPersonRenderer', (_ProtectedPersonRenderer,),
                {'person': person})


class _LinkedPersonRenderer(formalchemy.FieldRenderer):

    def render_readonly(self, **kwargs):
        return tags.link_to(str(self.raw_value),
                            self.request.route_url('person.edit', uuid=self.value))


def LinkedPersonRenderer(request):
    return type('LinkedPersonRenderer', (_LinkedPersonRenderer,),
                {'request': request})


class PasswordFieldRenderer(formalchemy.PasswordFieldRenderer):

    def render(self, **kwargs):
        return tags.password(self.name, value='', maxlength=self.length, **kwargs)


def passwords_match(value, field):
    if field.parent.confirm_password.value != value:
        raise formalchemy.ValidationError("Passwords do not match")
    return value


class PasswordField(formalchemy.Field):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('value', lambda x: x.password)
        kwargs.setdefault('renderer', PasswordFieldRenderer)
        kwargs.setdefault('validate', passwords_match)
        super(PasswordField, self).__init__(*args, **kwargs)

    def sync(self):
        if not self.is_readonly():
            password = self.renderer.deserialize()
            if password:
                auth.set_user_password(self.model, password)


class UserCrud(CrudView):

    mapped_class = edbob.User
    home_route = 'users'

    def fieldset(self, user):
        fs = self.make_fieldset(user)

        fs.append(PasswordField('password'))
        fs.append(formalchemy.Field('confirm_password',
                                    renderer=PasswordFieldRenderer))
        fs.append(RolesField('roles',
                             renderer=RolesFieldRenderer(self.request)))

        fs.configure(
            include=[
                fs.username,
                fs.person,
                fs.password.label("Set Password"),
                fs.confirm_password,
                fs.roles,
                ])

        if self.readonly:
            del fs.password
            del fs.confirm_password

        # if fs.edit and user.person:
        if isinstance(user, edbob.User) and user.person:
            fs.person.set(readonly=True,
                          renderer=LinkedPersonRenderer(self.request))

        return fs

    def validation_failed(self, fs):
        if not fs.edit and fs.person_uuid.value:
            fs.person.set(readonly=True,
                          renderer=ProtectedPersonRenderer(fs.person_uuid.value))


def includeme(config):

    config.add_route('users', '/users')
    config.add_view(UsersGrid, route_name='users',
                    renderer='/users/index.mako',
                    permission='users.list')

    config.add_route('user.create', '/users/new')
    config.add_view(UserCrud, attr='create', route_name='user.create',
                    renderer='/users/crud.mako',
                    permission='users.create')

    config.add_route('user.read', '/users/{uuid}')
    config.add_view(UserCrud, attr='read', route_name='user.read',
                    renderer='/users/crud.mako',
                    permission='users.read')

    config.add_route('user.update', '/users/{uuid}/edit')
    config.add_view(UserCrud, attr='update', route_name='user.update',
                    renderer='/users/crud.mako',
                    permission='users.update')

    config.add_route('user.delete', '/users/{uuid}/delete')
    config.add_view(UserCrud, attr='delete', route_name='user.delete',
                    permission='users.delete')



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.forms.formalchemy.fieldset`` -- FormAlchemy FieldSet
"""

import formalchemy

from edbob.pyramid import Session
from edbob.util import prettify


__all__ = ['FieldSet', 'make_fieldset']


class FieldSet(formalchemy.FieldSet):
    """
    Adds a little magic to the :class:`formalchemy.FieldSet` class.
    """

    prettify = staticmethod(prettify)

    def __init__(self, model, class_name=None, crud_title=None, url=None,
                 route_name=None, action_url='', home_url=None, **kwargs):
        super(FieldSet, self).__init__(model, **kwargs)
        self.class_name = class_name or self._original_cls.__name__.lower()
        self.crud_title = crud_title or prettify(self.class_name)
        self.edit = isinstance(model, self._original_cls)
        self.route_name = route_name or (self.class_name + 's')
        self.action_url = action_url
        self.home_url = home_url
        self.allow_continue = kwargs.pop('allow_continue', False)

    def get_display_text(self):
        return unicode(self.model)

    def render(self, **kwargs):
        kwargs.setdefault('class_', self.class_name)
        return super(FieldSet, self).render(**kwargs)


def make_fieldset(model, **kwargs):
    """
    Returns a :class:`FieldSet` equipped with the current scoped
    :class:`edbob.db.Session` instance (unless ``session`` is provided as a
    keyword argument).
    """

    kwargs.setdefault('session', Session())
    return FieldSet(model, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.grids.util`` -- Grid Utilities
"""

from sqlalchemy.orm.attributes import InstrumentedAttribute

from webhelpers.html import literal

from pyramid.response import Response

from edbob.pyramid.grids.search import SearchFormRenderer


def get_sort_config(name, request, **kwargs):
    """
    Returns a configuration dictionary for grid sorting.
    """

    # Initial config uses some default values.
    config = {
        'dir': 'asc',
        'per_page': 20,
        'page': 1,
        }

    # Override with defaults provided by caller.
    config.update(kwargs)

    # Override with values from GET/POST request and/or session.
    for key in config:
        full_key = name+'_'+key
        if request.params.get(key):
            value = request.params[key]
            config[key] = value
            request.session[full_key] = value
        elif request.session.get(full_key):
            value = request.session[full_key]
            config[key] = value

    return config


def get_sort_map(cls, names=None, **kwargs):
    """
    Convenience function which returns a sort map for ``cls``.

    If ``names`` is not specified, the map will include all "standard" fields
    present on the mapped class.  Otherwise, the map will be limited to only
    the fields which are named.

    All remaining ``kwargs`` are assumed to be sort map entries, and will be
    added to the map directly.
    """

    smap = {}
    if names is None:
        names = []
        for attr in cls.__dict__:
            obj = getattr(cls, attr)
            if isinstance(obj, InstrumentedAttribute):
                if obj.key != 'uuid':
                    names.append(obj.key)
    for name in names:
        smap[name] = sorter(getattr(cls, name))
    smap.update(kwargs)
    return smap


def render_grid(grid, search_form=None, **kwargs):
    """
    Convenience function to render ``grid`` (which should be a
    :class:`edbob.pyramid.grids.Grid` instance).

    This "usually" will return a dictionary to be used as context for rendering
    the final view template.

    However, if a partial grid is requested (or mandated), then the grid body
    will be rendered and a :class:`pyramid.response.Response` object will be
    returned instead.
    """

    if grid.partial_only or grid.request.params.get('partial'):
        return Response(body=grid.render(), content_type='text/html')
    kwargs['grid'] = literal(grid.render())
    if search_form:
        kwargs['search'] = SearchFormRenderer(search_form)
    return kwargs


def sort_query(query, config, sort_map, join_map={}):
    """
    Sorts ``query`` according to ``config`` and ``sort_map``.  ``join_map`` is
    used, if necessary, to join additional tables to the base query.  The
    sorted query is returned.
    """

    field = config.get('sort')
    if not field:
        return query
    joins = config.setdefault('joins', [])
    if field in join_map and field not in joins:
        query = join_map[field](query)
        joins.append(field)
    sort = sort_map[field]
    return sort(query, config['dir'])


def sorter(field):
    """
    Returns a function suitable for a sort map callable, with typical logic
    built in for sorting applied to ``field``.
    """

    return lambda q, d: q.order_by(getattr(field, d)())



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.commands`` -- Console Commands
"""

from __future__ import absolute_import

import sys
import argparse
import subprocess
import logging
import platform

import edbob
from edbob.util import entry_point_map, requires_impl


class ArgumentParser(argparse.ArgumentParser):
    """
    Customized version of ``argparse.ArgumentParser``, which overrides some of
    the argument parsing logic.  This is necessary for the application's
    primary command (:class:`Command` class); but is not used with
    :class:`Subcommand` derivatives.
    """

    def parse_args(self, args=None, namespace=None):
        args, argv = self.parse_known_args(args, namespace)
        args.argv = argv
        return args


class Command(edbob.Object):
    """
    The primary command for the application.  

    You should subclass this within your own application if you wish to
    leverage the command system provided by edbob.
    """
    
    name = 'edbob'
    version = edbob.__version__
    description = "Pythonic Software Framework"

    long_description = """
edbob is a Pythonic software framework.

Copyright (c) 2010-2012 Lance Edgar <lance@edbob.org>

This program comes with ABSOLUTELY NO WARRANTY.  This is free software,
and you are welcome to redistribute it under certain conditions.
See the file COPYING.txt for more information.
"""

    def __init__(self, **kwargs):
        edbob.Object.__init__(self, **kwargs)
        self.subcommands = entry_point_map('%s.commands' % self.name)

    def __str__(self):
        return str(self.name)

    def iter_subcommands(self):
        """
        Generator which yields associated :class:`Subcommand` classes, sorted
        by :attr:`Subcommand.name`.
        """

        for name in sorted(self.subcommands):
            yield self.subcommands[name]

    def print_help(self):
        """
        Prints help text for the primary command, including a list of available
        subcommands.
        """

        print """%(description)s

Usage: %(name)s [options] <command> [command-options]

Options:
  -c PATH, --config=PATH
                    Config path (may be specified more than once)
  -n, --no-init     Don't load config before executing command
  -d, --debug       Increase logging level to DEBUG
  -v, --verbose     Increase logging level to INFO
  -V, --version     Display program version and exit

Commands:""" % self

        for cmd in self.iter_subcommands():
            print "  %-16s  %s" % (cmd.name, cmd.description)

        print """
Try '%(name)s help <command>' for more help.""" % self

    def run(self, *args):
        """
        Parses ``args`` and executes the appropriate subcommand action
        accordingly (or displays help text).
        """

        parser = ArgumentParser(
            prog=self.name,
            description=self.description,
            add_help=False,
            )

        parser.add_argument('-c', '--config', action='append', dest='config_paths',
                            metavar='PATH')
        parser.add_argument('-d', '--debug', action='store_true', dest='debug')
        parser.add_argument('-n', '--no-init', action='store_true', default=False)
        parser.add_argument('-v', '--verbose', action='store_true', dest='verbose')
        parser.add_argument('-V', '--version', action='version',
                            version="%%(prog)s %s" % self.version)
        parser.add_argument('command', nargs='*')

        # Parse args and determind subcommand.
        args = parser.parse_args(list(args))
        if not args or not args.command:
            self.print_help()
            return

        # Show (sub)command help if so instructed, or unknown subcommand.
        cmd = args.command.pop(0)
        if cmd == 'help':
            if len(args.command) != 1:
                self.print_help()
                return
            cmd = args.command[0]
            if cmd not in self.subcommands:
                self.print_help()
                return
            cmd = self.subcommands[cmd](parent=self)
            cmd.parser.print_help()
            return
        elif cmd not in self.subcommands:
            self.print_help()
            return

        # Basic logging should be established before init()ing.

        # Use root logger if setting logging flags.
        log = logging.getLogger()
        edbob.basic_logging()
        if args.verbose:
            log.setLevel(logging.INFO)
        if args.debug:
            log.setLevel(logging.DEBUG)

        # Initialize everything...
        if not args.no_init:
            edbob.init(self.name, *(args.config_paths or []))

            # Command line logging flags should override config.
            if args.verbose:
                log.setLevel(logging.INFO)
            if args.debug:
                log.setLevel(logging.DEBUG)
        
        # And finally, do something of real value...
        cmd = self.subcommands[cmd](parent=self)
        cmd._run(*(args.command + args.argv))


class Subcommand(edbob.Object):
    """
    Base class for application "subcommands."  You'll want to derive from this
    class as needed to implement most of your application's command logic.
    """

    def __init__(self, **kwargs):
        edbob.Object.__init__(self, **kwargs)
        self.parser = argparse.ArgumentParser(
            prog='%s %s' % (self.parent, self.name),
            description=self.description,
            )
        self.add_parser_args(self.parser)

    @property
    @requires_impl(is_property=True)
    def name(self):
        """
        The name of the subcommand.

        .. note::
           The subcommand name should ideally be limited to 12 characters in
           order to preserve formatting consistency when displaying help text.
        """
        pass

    @property
    @requires_impl(is_property=True)
    def description(self):
        """
        The description for the subcommand.
        """
        pass

    def __repr__(self):
        return "<Subcommand: %s>" % self.name

    def add_parser_args(self, parser):
        """
        If your subcommand accepts optional (or positional) arguments, you
        should override this and add the possible arguments directly to
        ``parser`` (which is a :class:`argparse.ArgumentParser` instance) via
        its ``add_argument()`` method.
        """
        pass
            
    def _run(self, *args):
        args = self.parser.parse_args(list(args))
        return self.run(args)

    @requires_impl()
    def run(self, args):
        """
        Runs the subcommand.  You must override this method within your
        subclass.  ``args`` will be a :class:`argparse.Namespace` object
        containing all parsed arguments found on the original command line
        executed by the user.
        """
        pass


class DatabaseCommand(Subcommand):
    """
    Provides tools for managing an ``edbob`` database; called as ``edbob db``.

    This command requires additional arguments; see ``edbob help db`` for more
    information.
    """

    name = 'db'
    description = "Tools for managing edbob databases"

    def add_parser_args(self, parser):
        parser.add_argument('-D', '--database', metavar='URL',
                            help="Database engine (default is edbob.db.engine)")
        # parser.add_argument('command', choices=['upgrade', 'extensions', 'activate', 'deactivate'],
        #                     help="Command to execute against database")
        subparsers = parser.add_subparsers(title='subcommands')

        extensions = subparsers.add_parser('extensions',
                                           help="Display current extension status for the database")
        extensions.set_defaults(func=self.extensions)

        activate = subparsers.add_parser('activate',
                                         help="Activate an extension within the database")
        activate.add_argument('extension', help="Name of extension to activate")
        activate.set_defaults(func=self.activate)

        deactivate = subparsers.add_parser('deactivate',
                                           help="Deactivate an extension within the database")
        deactivate.add_argument('extension', help="Name of extension to deactivate")
        deactivate.set_defaults(func=self.deactivate)

    def activate(self, engine, args):
        from edbob.db.extensions import (
            available_extensions,
            extension_active,
            activate_extension,
            )

        if args.extension in available_extensions():
            if not extension_active(args.extension, engine):
                activate_extension(args.extension, engine)
                print "Activated extension '%s' in database:" % args.extension
                print '  %s' % engine.url
            else:
                print >> sys.stderr, "Extension already active: %s" % args.extension
        else:
            print >> sys.stderr, "Extension unknown: %s" % args.extension

    def deactivate(self, engine, args):
        from edbob.db.extensions import (
            available_extensions,
            extension_active,
            deactivate_extension,
            )

        if args.extension in available_extensions():
            if extension_active(args.extension, engine):
                deactivate_extension(args.extension, engine)
                print "Deactivated extension '%s' in database:" % args.extension
                print '  %s' % engine.url
            else:
                print >> sys.stderr, "Extension already inactive: %s" % args.extension
        else:
            print >> sys.stderr, "Extension unknown: %s" % args.extension

    def extensions(self, engine, args):
        from edbob.db.extensions import (
            available_extensions,
            extension_active,
            )

        print "Extensions for database:"
        print '  %s' % engine.url
        print ''
        print " Name            Active?"
        print "------------------------"
        for name in sorted(available_extensions()):
            print " %-16s  %s" % (
                name, 'Yes' if extension_active(name, engine) else 'No')
        print ''
        print "Use 'edbob db [de]activate <extension>' to change."

    def run(self, args):
        if args.database:
            from sqlalchemy import create_engine
            from sqlalchemy.exc import ArgumentError
            try:
                engine = create_engine(args.database)
            except ArgumentError, err:
                print err
                return
        else:
            from edbob.db import engine
        if not engine:
            print >> sys.stderr, "Database not configured; please change that or specify -D URL"
            return
        args.func(engine, args)


# class ExtensionsCommand(RattailCommand):
#     """
#     Displays the currently-installed (available) extensions, and whether or not
#     each has been activated within the configured database.  Called as
#     ``rattail extensions``.
#     """

#     short_description = "Displays available extensions and their statuses"

#     def __call__(self, options, **args):
#         from sqlalchemy.exc import OperationalError
#         from rattail import engine
#         from rattail.db.util import core_schema_installed
#         from rattail.db.ext import get_available_extensions, extension_active

#         try:
#             engine.connect()
#         except OperationalError, e:
#             print >> sys.stderr, "Cannot connect to database:"
#             print >> sys.stderr, engine.url
#             print >> sys.stderr, e[0].strip()
#             return

#         if not core_schema_installed():
#             print >> sys.stderr, "Database lacks core schema:"
#             print >> sys.stderr, engine.url
#             return

#         extensions = get_available_extensions()
#         print ''
#         print ' %-25s %-10s' % ("Extension", "Active?")
#         print '-' * 35
#         for name in sorted(extensions):
#             active = 'Y' if extension_active(name) else 'N'
#             print ' %-28s %s' % (name, active)
#         print ''
#         print "Use 'rattail {activate|deactivate} EXTENSION' to change."

        
class FileMonitorCommand(Subcommand):
    """
    Interacts with the file monitor service; called as ``edbob filemon``.  This
    command expects a subcommand; one of the following:

    * ``edbob filemon start``
    * ``edbob filemon stop``

    On Windows platforms, the following additional subcommands are available:

    * ``edbob filemon install``
    * ``edbob filemon uninstall``

    .. note::
       The Windows Vista family of operating systems requires you to launch
       ``cmd.exe`` as an Administrator in order to have sufficient rights to
       run the above commands.

       .. todo::
          Verify the previous statement...  (Maybe test with/out UAC?)

    See :doc:`howto.use_filemon` for more information.
    """

    name = 'filemon'
    description = "Manage the file monitor service"

    appname = 'edbob'

    def add_parser_args(self, parser):
        subparsers = parser.add_subparsers(title='subcommands')

        start = subparsers.add_parser('start', help="Start service")
        start.set_defaults(subcommand='start')
        stop = subparsers.add_parser('stop', help="Stop service")
        stop.set_defaults(subcommand='stop')

        if sys.platform == 'win32':
            install = subparsers.add_parser('install', help="Install service")
            install.set_defaults(subcommand='install')
            install.add_argument('-a', '--auto-start', action='store_true',
                                 help="Configure service to start automatically")
            remove = subparsers.add_parser('remove', help="Uninstall (remove) service")
            remove.set_defaults(subcommand='remove')
            uninstall = subparsers.add_parser('uninstall', help="Uninstall (remove) service")
            uninstall.set_defaults(subcommand='remove')

        else:
            parser.add_argument('-D', '--dont-daemonize',
                                action='store_false', dest='daemonize',
                                help="Don't daemonize when starting")

    def get_win32_module(self):
        from edbob.filemon import win32
        return win32

    def get_win32_service(self):
        from edbob.filemon.win32 import FileMonitorService
        return FileMonitorService

    def get_win32_service_name(self):
        service = self.get_win32_service()
        return service._svc_name_

    def run(self, args):
        if sys.platform == 'linux2':
            from edbob.filemon import linux as filemon

            if args.subcommand == 'start':
                filemon.start_daemon(self.appname, daemonize=args.daemonize)

            elif args.subcommand == 'stop':
                filemon.stop_daemon()

        elif sys.platform == 'win32':
            from edbob import win32

            filemon = self.get_win32_module()

            # Execute typical service command.
            options = []
            if args.subcommand == 'install' and args.auto_start:
                options = ['--startup', 'auto']
            win32.execute_service_command(filemon, args.subcommand, *options)

            # If installing auto-start service on Windows 7, we should update
            # its startup type to be "Automatic (Delayed Start)".
            if args.subcommand == 'install' and args.auto_start:
                if platform.release() == '7':
                    name = self.get_win32_service_name()
                    win32.delayed_auto_start_service(name)

        else:
            print "Sorry, file monitor is not supported on platform %s." % sys.platform


class ShellCommand(Subcommand):
    """
    Launches a Python shell (of your choice) with ``edbob`` pre-loaded; called
    as ``edbob shell``.

    You can define the shell within your config file (otherwise ``python`` is
    assumed)::

    .. highlight:: ini

       [edbob]
       shell.python = ipython
    """

    name = 'shell'
    description = "Launch Python shell with edbob pre-loaded"

    def run(self, args):
        code = ['import edbob']
        if edbob.inited:
            code.append("edbob.init('edbob', %s, shell=True)" %
                        edbob.config.paths_loaded)
        code.append('from edbob import *')
        code = '; '.join(code)
        print "edbob v%s launching Python shell...\n" % edbob.__version__
        python = 'python'
        if edbob.inited:
            python = edbob.config.get('edbob', 'shell.python', default=python)
        proc = subprocess.Popen([python, '-i', '-c', code])
        while True:
            try:
                proc.wait()
            except KeyboardInterrupt:
                pass
            else:
                break


class UuidCommand(Subcommand):
    """
    Command for generating an UUID; called as ``edbob uuid``.

    If the ``--gui`` option is specified, this command launches a small
    wxPython application for generating UUIDs; otherwise a single UUID will be
    printed to the console.
    """

    name = 'uuid'
    description = "Generate an universally-unique identifier"

    def add_parser_args(self, parser):
        parser.add_argument('--gui', action='store_true',
                            help="Display graphical interface")

    def run(self, args):
        if args.gui:
            from edbob.wx.GenerateUuid import main
            main()
        else:
            print edbob.get_uuid()


def main(*args):
    """
    The primary entry point for the edbob command system.

    .. note::
       This entry point is really for ``edbob`` proper.  Your application
       should provide its own command entry point which leverages *your*
       :class:`Command` subclass instead of using this entry point (which uses
       :class:`Command` directly).

       There's not much involved in doing so; see the source code for
       implementation details.
    """

    if args:
        args = list(args)
    else:
        args = sys.argv[1:]

    cmd = Command()
    cmd.run(*args)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.sqlalchemy`` -- SQLAlchemy Stuff
"""

from __future__ import absolute_import

from sqlalchemy import Table, Column, String
from sqlalchemy import engine_from_config as _engine_from_config

from edbob.core import get_uuid
from edbob.modules import load_spec
from edbob.time import utc_time


__all__ = ['getset_factory', 'table_with_uuid', 'current_time']


def current_time(context):
    """
    This function may be provided to the ``default`` parameter of a
    :class:`sqlalchemy.Column` class definition.  Doing so will ensure the
    column's default value will be the current time in UTC.
    """

    return utc_time(naive=True)


def engine_from_config(config, prefix='sqlalchemy.', **kwargs):
    """
    Slightly enhanced version of the :func:`sqlalchemy.engine_from_config()`
    function.  This version is aware of the ``poolclass`` configuration
    parameter, and will coerce it via :func:`edbob.load_spec()`.

    Note that if a pool class is specified, the class should be represented
    using the "spec" format and *not* pure dotted path notation, e.g.:

    Correct::

       [edbob.db]
       default.poolclass = sqlqlchemy.pool:NullPool

    Incorrect::

       [edbob.db]
       default.poolclass = sqlalchemy.pool.NullPool
    """

    if config.has_key(prefix + 'poolclass'):
        config[prefix + 'poolclass'] = load_spec(config[prefix + 'poolclass'])
    return _engine_from_config(config, prefix=prefix, **kwargs)


def getset_factory(collection_class, proxy):
    """
    Get/set factory for SQLAlchemy association proxy attributes.
    """

    def getter(obj):
        if obj is None:
            return None
        return getattr(obj, proxy.value_attr)

    def setter(obj, val):
        setattr(obj, proxy.value_attr, val)

    return getter, setter


def table_with_uuid(name, metadata, *args, **kwargs):
    """
    .. highlight:: python

    Convenience function to abstract the addition of the ``uuid`` column to a
    new table.  Can be used to replace this::

       Table(
           'things', metadata,
           Column('uuid', String(32), primary_key=True, default=get_uuid),
           Column('name', String(50)),
           )

    ...with this::

        table_with_uuid(
            'things', metadata,
            Column('name', String(50)),
            )
    """

    return Table(name, metadata,
                 Column('uuid', String(32), primary_key=True, default=get_uuid),
                 *args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.forms.formalchemy`` -- FormAlchemy Interface
"""

from __future__ import absolute_import

import datetime

from pyramid.renderers import render
from webhelpers.html.tags import literal

import formalchemy
from formalchemy.validators import accepts_none

import edbob
from edbob.lib import pretty
from edbob.pyramid import Session
from edbob.time import localize
from edbob.util import requires_impl

from edbob.pyramid.forms.formalchemy.fieldset import *
from edbob.pyramid.forms.formalchemy.fields import *
from edbob.pyramid.forms.formalchemy.renderers import *


__all__ = ['ChildGridField', 'PropertyField', 'EnumFieldRenderer',
           'PrettyDateTimeFieldRenderer', 'AutocompleteFieldRenderer',
           'FieldSet', 'make_fieldset', 'required', 'pretty_datetime',
           'AssociationProxyField', 'YesNoFieldRenderer']


class TemplateEngine(formalchemy.templates.TemplateEngine):
    """
    Mako template engine for FormAlchemy.
    """

    def render(self, template, prefix='/forms/', suffix='.mako', **kwargs):
        template = ''.join((prefix, template, suffix))
        return render(template, kwargs)


# Make our TemplateEngine the default.
engine = TemplateEngine()
formalchemy.config.engine = engine


class AlchemyForm(edbob.Object):
    """
    Form to contain a :class:`formalchemy.FieldSet` instance.
    """

    create_label = "Create"
    update_label = "Update"

    allow_successive_creates = False

    def __init__(self, request, fieldset, **kwargs):
        edbob.Object.__init__(self, **kwargs)
        self.request = request
        self.fieldset = fieldset

    def _get_readonly(self):
        return self.fieldset.readonly

    def _set_readonly(self, val):
        self.fieldset.readonly = val

    readonly = property(_get_readonly, _set_readonly)

    @property
    def successive_create_label(self):
        return "%s and continue" % self.create_label

    def render(self, **kwargs):
        kwargs['form'] = self
        if self.readonly:
            template = '/forms/form_readonly.mako'
        else:
            template = '/forms/form.mako'
        return render(template, kwargs)

    def save(self):
        self.fieldset.sync()
        Session.flush()

    def validate(self):
        self.fieldset.rebind(data=self.request.params)
        return self.fieldset.validate()


class ChildGridField(formalchemy.Field):
    """
    Convenience class for including a child grid within a fieldset as a
    read-only field.
    """

    def __init__(self, name, value, *args, **kwargs):
        super(ChildGridField, self).__init__(name, *args, **kwargs)
        self.set(value=value)
        self.set(readonly=True)


class PropertyField(formalchemy.Field):
    """
    Convenience class for fields which simply involve a read-only property
    value.
    """

    def __init__(self, name, attr=None, *args, **kwargs):
        super(PropertyField, self).__init__(name, *args, **kwargs)
        if not attr:
            attr = name
        self.set(value=lambda x: getattr(x, attr))
        self.set(readonly=True)


@accepts_none
def required(value, field=None):
    if value is None or value == '':
        msg = "Please provide a value"
        if field:
            msg = "You must provide a value for %s" % field.label()
        raise formalchemy.ValidationError(msg)


def pretty_datetime(value, from_='local', to='local'):
    """
    Formats a ``datetime.datetime`` instance and returns a "pretty"
    human-readable string from it, e.g. "42 minutes ago".  ``value`` is
    rendered directly as a string if no date/time can be parsed from it.
    """

    if not isinstance(value, datetime.datetime):
        return str(value) if value else ''
    if not value.tzinfo:
        value = localize(value, from_=from_, to=to)
    return literal('<span title="%s">%s</span>' % (
            value.strftime('%Y-%m-%d %H:%M:%S %Z%z'),
            pretty.date(value)))    


def PrettyDateTimeFieldRenderer(from_='local', to='local'):

    class PrettyDateTimeFieldRenderer(formalchemy.fields.DateTimeFieldRenderer):

        def render_readonly(self, **kwargs):
            return pretty_datetime(self.raw_value, from_=from_, to=to)

    return PrettyDateTimeFieldRenderer


class DateTimeFieldRenderer(formalchemy.fields.DateTimeFieldRenderer):
    """
    Leverages edbob time system to coerce timestamp to local time zone before
    displaying it.
    """

    def render_readonly(self, **kwargs):
        value = self.raw_value
        if isinstance(value, datetime.datetime):
            value = edbob.local_time(value)
            return value.strftime(self.format)
        return ''

FieldSet.default_renderers[formalchemy.types.DateTime] = DateTimeFieldRenderer



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid`` -- Pyramid Framework
"""

from sqlalchemy.orm import sessionmaker, scoped_session
from zope.sqlalchemy import ZopeTransactionExtension

import edbob


__all__ = ['Session']

Session = scoped_session(sessionmaker())


def includeme(config):
    """
    Adds ``edbob``-specific features to the application.  Currently this does
    two things:

    It adds a ``ZopeTransactionExtension`` instance as an extension to the
    SQLAlchemy scoped ``Session`` class.  This is necessary for most view code
    that ships with ``edbob``, so you will most likely need to specify
    ``config.include('edbob.pyramid')`` somewhere in your app config (i.e. your
    ``main()`` function).

    The other thing added is the ``edbob`` static view for CSS files etc.
    """

    # Configure Beaker session.
    config.include('pyramid_beaker')

    # Bring in transaction manager.
    config.include('pyramid_tm')

    # Configure SQLAlchemy session.
    Session.configure(bind=edbob.engine)
    Session.configure(extension=ZopeTransactionExtension())

    # Configure user authentication / authorization.
    from pyramid.authentication import SessionAuthenticationPolicy
    config.set_authentication_policy(SessionAuthenticationPolicy())
    from edbob.pyramid.auth import EdbobAuthorizationPolicy
    config.set_authorization_policy(EdbobAuthorizationPolicy())

    # Add forbidden view.
    config.add_forbidden_view('edbob.pyramid.views.forbidden')

    # Add static views.
    config.include('edbob.pyramid.static')

    # Add subscriber hooks.
    config.include('edbob.pyramid.subscribers')



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views`` -- Views
"""

from pyramid.httpexceptions import HTTPFound
from pyramid.security import authenticated_userid

from webhelpers.html import literal
from webhelpers.html.tags import link_to

from edbob.pyramid.views.core import *
from edbob.pyramid.views.grids import *
from edbob.pyramid.views.crud import *
from edbob.pyramid.views.autocomplete import *
# from edbob.pyramid.views.form import *


def forbidden(request):
    """
    The forbidden view.  This is triggered whenever access rights are denied
    for an otherwise-appropriate view.
    """

    msg = literal("You do not have permission to do that.")
    if not authenticated_userid(request):
        msg += literal("&nbsp; (Perhaps you should %s?)" %
                       link_to("log in", request.route_url('login')))
    request.session.flash(msg, allow_duplicate=False)

    url = request.referer
    if not url or url == request.current_route_url():
        url = request.route_url('home')
    return HTTPFound(location=url)


def includeme(config):
    config.include('edbob.pyramid.views.auth')
    config.include('edbob.pyramid.views.people')
    config.include('edbob.pyramid.views.progress')
    config.include('edbob.pyramid.views.users')



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.auth`` -- Authentication & Authorization
"""

from zope.interface import implementer

from pyramid.interfaces import IAuthorizationPolicy
from pyramid.security import Everyone, Authenticated

import edbob
from edbob.db.auth import has_permission
from edbob.pyramid import Session


# def groupfinder(userid, request):
#     q = Session.query(edbob.UserRole)
#     q = q.filter(edbob.UserRole.user_uuid == userid)
#     return [x.role_uuid for x in q]


@implementer(IAuthorizationPolicy)
class EdbobAuthorizationPolicy(object):

    def permits(self, context, principals, permission):
        for userid in principals:
            if userid not in (Everyone, Authenticated):
                user = Session.query(edbob.User).get(userid)
                assert user
                return has_permission(user, permission)
        if Everyone in principals:
            return has_permission(None, permission, session=Session())
        return False

    def principals_allowed_by_permission(self, context, permission):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.db`` -- Database Framework
"""

from __future__ import absolute_import

from sqlalchemy import MetaData
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base

import edbob
from edbob.sqlalchemy import engine_from_config


__all__ = ['engines', 'engine', 'Session', 'get_setting', 'save_setting']

engines = None
engine = None
Session = sessionmaker()
Base = declarative_base(cls=edbob.Object)


def init(config):
    """
    Initializes the database connection(s); called by :func:`edbob.init()` if
    config includes something like::

       .. highlight:: ini

       [edbob]
       init = ['edbob.db']

       [edbob.db]
       sqlalchemy.urls = {
               'default':       'postgresql://user:pass@localhost/edbob,
               }

    This function reads connection info from ``config`` and builds a dictionary
    or :class:`sqlalchemy.Engine` instances accordingly.  It also extends the
    root ``edbob`` namespace with the ORM classes (:class:`edbob.Person`,
    :class:`edbob.User`, etc.), as well as a few other things
    (e.g. :attr:`edbob.engine`, :attr:`edbob.Session`, :attr:`edbob.metadata`).
    """

    import edbob.db
    from edbob.db import model
    from edbob.db import enum
    from edbob.db.extensions import extend_framework

    global engines, engine

    keys = config.get('edbob.db', 'keys')
    if keys:
        keys = keys.split(',')
    else:
        keys = ['default']

    engines = {}
    cfg = config.get_dict('edbob.db')
    for key in keys:
        key = key.strip()
        try:
            engines[key] = engine_from_config(cfg, '%s.' % key)
        except KeyError:
            if key == 'default':
                try:
                    engines[key] = engine_from_config(cfg, 'sqlalchemy.')
                except KeyError:
                    pass

    engine = engines.get('default')
    if engine:
        Session.configure(bind=engine)
    
    extend_framework()

    edbob.graft(edbob, edbob.db)
    edbob.graft(edbob, model)
    edbob.graft(edbob, enum)


def get_setting(name, session=None):
    """
    Returns a setting from the database.
    """

    _session = session
    if not session:
        session = Session()
    setting = session.query(edbob.Setting).get(name)
    if setting:
        setting = setting.value
    if not _session:
        session.close()
    return setting


def save_setting(name, value, session=None):
    """
    Saves a setting to the database.
    """

    _session = session
    if not session:
        session = Session()
    setting = session.query(edbob.Setting).get(name)
    if not setting:
        setting = edbob.Setting(name=name)
        session.add(setting)
    setting.value = value
    if not _session:
        session.commit()
        session.close()


def get_core_metadata():
    """
    Returns a :class:`sqlalchemy.MetaData` instance containing only those
    :class:`sqlalchemy.Table`s which are part of the core ``edbob`` schema.
    """

    from edbob.db import model

    meta = MetaData()
    for name in model.__all__:
        if name != 'Base':
            obj = getattr(model, name)
            if isinstance(obj, type) and issubclass(obj, model.Base):
                obj.__table__.tometadata(meta)
    return meta


def needs_session(func):
    """
    Decorator which adds helpful session handling.
    """

    def wrapped(*args, **kwargs):
        session = kwargs.pop('session', None)
        _orig_session = session
        if not session:
            session = Session()
        res = func(session, *args, **kwargs)
        if not _orig_session:
            session.commit()
            session.close()
        return res

    return wrapped



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.progress`` -- Progress Views
"""

from edbob.pyramid.progress import get_progress_session


def progress(request):
    key = request.matchdict['key']
    session = get_progress_session(request.session, key)
    if session.get('complete'):
        request.session.flash(session.get('success_msg', "The process has completed successfully."))
    elif session.get('error'):
        request.session.flash(session.get('error_msg', "An unspecified error occurred."), 'error')
    return session


def cancel(request):
    key = request.matchdict['key']
    session = get_progress_session(request.session, key)
    session.clear()
    session['canceled'] = True
    session.save()
    msg = request.params.get('cancel_msg', "The operation was canceled.")
    request.session.flash(msg)
    return {}


def includeme(config):
    config.add_route('progress', '/progress/{key}')
    config.add_view(progress, route_name='progress', renderer='json')

    config.add_route('progress.cancel', '/progress/{key}/cancel')
    config.add_view(cancel, route_name='progress.cancel', renderer='json')



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.core`` -- Core Stuff
"""


import logging
import uuid


__all__ = ['Object', 'basic_logging', 'get_uuid', 'graft']


class Object(object):
    """
    Generic base class which provides a common ancestor, and some other
    conveniences.
    """

    def __init__(self, **kwargs):
        """
        Constructor.  All keyword arguments are assumed to be attribute names
        and are assigned directly to the new ``Object`` instance.
        """

        for key in kwargs:
            setattr(self, key, kwargs[key])

    def __getitem__(self, key):
        """
        Allows dict-like access to the object's attributes.
        """

        if hasattr(self, key):
            return getattr(self, key)

    def __str__(self):
        """
        Leverage :meth:`__unicode__()` method if it exists; otherwise fall back
        to ``repr(self)``.
        """

        if hasattr(self, '__unicode__'):
            return str(unicode(self))
        return repr(self)


def basic_logging():
    """
    Does some basic configuration on the root logger.

    .. note::
       This only enables console output at this point; it is assumed that if
       you intend to "truly" configure logging that you will be using a proper
       config file and calling :func:`edbob.init()`.
    """

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(
            '%(levelname)-5.5s [%(name)s][%(threadName)s] %(message)s'))
    root = logging.getLogger()
    root.addHandler(handler)


def get_uuid():
    """
    Generates a universally-unique identifier and returns its 32-character hex
    value.
    """

    return uuid.uuid1().hex


def graft(target, source, names=None):
    """
    Adds names to the ``target`` namespace, copying each from ``source``.

    If ``names`` is provided, it can be a string if adding only one thing;
    otherwise it should be a list of strings.  If it is not provided, then
    everything from ``source`` will be grafted.

    .. note::
       If "everything" is to be grafted (i.e. ``names is None``), then
       ``source.__all__`` will be consulted if available.  If it is not, then
       ``dir(source)`` will be used instead.
    """

    if names is None:
        if hasattr(source, '__all__'):
            names = source.__all__
        else:
            names = dir(source)
    elif isinstance(names, basestring):
        names = [names]

    for name in names:
        if hasattr(source, name):
            setattr(target, name, getattr(source, name))
        else:
            setattr(target, name, source.get(name))
        if not hasattr(target, '__all__'):
            target.__all__ = []
        target.__all__.append(name)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.filemon.win32`` -- File Monitoring Service for Windows
"""

import os.path
import sys
import Queue
import logging
import threading

import edbob
from edbob.errors import email_exception
from edbob.filemon import get_monitor_profiles
from edbob.win32 import Service, file_is_free

if sys.platform == 'win32': # docs should build for everyone
    import win32api
    import win32con
    import win32event
    import win32file
    import win32service
    import win32serviceutil
    import winnt


log = logging.getLogger(__name__)


class FileMonitorService(Service):
    """
    Implements edbob's file monitor Windows service.
    """

    _svc_name_ = 'EdbobFileMonitor'
    _svc_display_name_ = "Edbob : File Monitoring Service"
    _svc_description_ = ("Monitors one or more folders for incoming files, "
                         "and performs configured actions as new files arrive.")

    def Initialize(self):
        """
        Service initialization.
        """

        if not Service.Initialize(self):
            return False

        # Read monitor profile(s) from config.
        self.monitored = get_monitor_profiles(self.appname)

        # Make sure we have something to do.
        if not self.monitored:
            return False

        # Create monitor and action threads for each profile.
        for key, profile in self.monitored.iteritems():

            # Create a file queue for the profile.
            queue = Queue.Queue()

            # Create a monitor thread for each folder in profile.
            for i, path in enumerate(profile.dirs, 1):
                name = 'monitor-%s-%u' % (key, i)
                log.debug("Initialize: Starting '%s' thread for folder: %s" %
                          (name, path))
                thread = threading.Thread(
                    target=monitor_files,
                    name=name,
                    args=(queue, path, profile))
                thread.daemon = True
                thread.start()

            # Create an action thread for the profile.
            name = 'actions-%s' % key
            log.debug("Initialize: Starting '%s' thread" % name)
            thread = threading.Thread(
                target=perform_actions,
                name=name,
                args=(queue, profile))
            thread.daemon = True
            thread.start()

        return True

    
def monitor_files(queue, path, profile):
    """
    Callable target for file monitor threads.
    """

    hDir = win32file.CreateFile(
        path,
        winnt.FILE_LIST_DIRECTORY,
        win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE,
        None,
        win32con.OPEN_EXISTING,
        win32con.FILE_FLAG_BACKUP_SEMANTICS,
        None)

    if hDir == win32file.INVALID_HANDLE_VALUE:
        log.warning("monitor_files: Can't open directory with CreateFile(): %s" % path)
        return

    while True:
        results = win32file.ReadDirectoryChangesW(
            hDir,
            1024,
            False,
            win32con.FILE_NOTIFY_CHANGE_FILE_NAME)

        log.debug("monitor_files: ReadDirectoryChangesW() results: %s" % results)
        for action, fname in results:
            fpath = os.path.join(path, fname)
            if action in (winnt.FILE_ACTION_ADDED,
                          winnt.FILE_ACTION_RENAMED_NEW_NAME):
                log.debug("monitor_files: Queueing '%s' file: %s" %
                          (profile.key, fpath))
                queue.put(fpath)


def perform_actions(queue, profile):
    """
    Callable target for action threads.
    """

    while True:

        try:
            path = queue.get_nowait()
        except Queue.Empty:
            pass
        else:

            while not file_is_free(path):
                win32api.Sleep(0)

            for spec, func, args in profile.actions:

                log.info("perform_actions: Calling function '%s' on file: %s" %
                         (spec, path))

                try:
                    func(path, *args)

                except:
                    log.exception("perform_actions: An exception occurred "
                                  "while processing file: %s" % path)
                    email_exception()

                    # This file probably shouldn't be processed any further.
                    break


if __name__ == '__main__':
    win32serviceutil.HandleCommandLine(FileMonitorService)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.errors`` -- Error Alert Emails
"""

import os.path
import sys
import socket
import logging
from traceback import format_exception
from cStringIO import StringIO

import edbob
from edbob.files import resource_path
from edbob.mail import sendmail_with_config


log = logging.getLogger(__name__)


def init(config):
    """
    Creates a system-wide exception hook which logs exceptions and emails them
    to configured recipient(s).
    """

    def excepthook(type, value, traceback):
        email_exception(type, value, traceback)
        sys.__excepthook__(type, value, traceback)

    sys.excepthook = excepthook


def email_exception(type=None, value=None, traceback=None):
    """
    Sends an email containing a traceback to the configured recipient(s).
    """

    if not (type and value and traceback):
        type, value, traceback = sys.exc_info()

    hostname = socket.gethostname()
    traceback = ''.join(format_exception(type, value, traceback))
    traceback = traceback.strip()
    data = {
        'host_name':    hostname,
        'host_ip':      socket.gethostbyname(hostname),
        'host_time':    edbob.local_time(),
        'traceback':    traceback,
        }

    body, ctype = render_exception(data)
    ctype = edbob.config.get('edbob.errors', 'content_type', default=ctype)
    sendmail_with_config('errors', body, content_type=ctype)


def render_exception(data):
    """
    Renders the exception data using a Mako template if one is configured;
    otherwise as a simple string.
    """

    template = edbob.config.get('edbob.errors', 'template')
    if template:
        template = resource_path(template)
        if os.path.exists(template):

            # Assume Mako template; render and return.
            from mako.template import Template
            template = Template(filename=template)
            return template.render(**data), 'text/html'

    # If not a Mako template, return regular text with substitutions.
    body = StringIO()
    data['host_time'] = data['host_time'].strftime('%Y-%m-%d %H:%M:%S %Z%z')

    body.write("""\
An unhandled exception occurred.

Machine Name:  %(host_name)s (%(host_ip)s)

Machine Time:  %(host_time)s

%(traceback)s
""" % data)

    b = body.getvalue()
    body.close()

    return b, 'text/plain'



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.configuration`` -- Configuration
"""

import os
import os.path
import sys
import ConfigParser
import logging
import logging.config

import edbob
from edbob import exceptions


__all__ = ['AppConfigParser']

log = logging.getLogger(__name__)


class AppConfigParser(ConfigParser.SafeConfigParser):
    """
    Subclass of ``ConfigParser.SafeConfigParser``, with some conveniences
    added.
    """

    def __init__(self, appname, *args, **kwargs):
        ConfigParser.SafeConfigParser.__init__(self, *args, **kwargs)
        self.appname = appname
        self.paths_attempted = []
        self.paths_loaded = []

    def clear(self):
        """
        Completely clears the contents of the config instance.
        """

        for section in self.sections():
            self.remove_section(section)
        del self.paths_attempted[:]
        del self.paths_loaded[:]

    def configure_logging(self):
        """
        Saves the current (possibly cascaded) configuration to a temporary
        file, and passes that to ``logging.config.fileConfig()``.
        """

        if self.getboolean('edbob', 'basic_logging', default=False):
            edbob.basic_logging()
        if self.getboolean('edbob', 'configure_logging', default=False):
            path = edbob.temp_path(suffix='.conf')
            self.save(path)
            try:
                logging.config.fileConfig(path, disable_existing_loggers=False)
            except ConfigParser.NoSectionError:
                pass
            os.remove(path)
        log.debug("Configured logging")

    def get(self, section, option, raw=False, vars=None, default=None):
        """
        Overridden version of ``ConfigParser.SafeConfigParser.get()``; this one
        adds the ``default`` keyword parameter and will return it instead of
        raising an error when the option doesn't exist.
        """

        if self.has_option(section, option):
            return ConfigParser.SafeConfigParser.get(self, section, option, raw, vars)
        return default

    def getboolean(self, section, option, default=None):
        """
        Overriddes base class method to allow for a default.
        """

        try:
            val = ConfigParser.SafeConfigParser.getboolean(self, section, option)
        except AttributeError:
            return default
        return val

    def get_dict(self, section):
        """
        Convenience method which returns a dictionary of options contained
        within the given section.
        """

        d = {}
        for opt in self.options(section):
            d[opt] = self.get(section, opt)
        return d

    def get_user_dir(self, create=False):
        """
        Returns a path to the "preferred" user-level folder, in which
        additional config files (etc.) may be placed as needed.  This
        essentially returns a platform-specific variation of ``~/.appname``.

        If ``create`` is ``True``, then the folder will be created if it does
        not already exist.
        """

        path = os.path.expanduser('~/.%s' % self.appname)
        if sys.platform == 'win32':
            from win32com.shell import shell, shellcon
            path = os.path.join(
                shell.SHGetSpecialFolderPath(0, shellcon.CSIDL_APPDATA),
                self.appname)
        if create and not os.path.exists(path):
            os.mkdir(path)
        return path

    def get_user_file(self, filename, create=False):
        """
        Returns a full path to a user-level config file location.  This is
        obtained by first calling :meth:`get_user_dir()` and then joining the
        result with ``filename``.

        The ``create`` argument will be passed directly to
        :meth:`get_user_dir()`, and may be used to ensure the user-level folder
        exists.
        """

        return os.path.join(self.get_user_dir(create=create), filename)

    def options(self, section):
        """
        Overridden version of ``ConfigParser.SafeConfigParser.options()``.
        This one doesn't raise an error if ``section`` doesn't exist, but
        instead returns an empty list.
        """

        if not self.has_section(section):
            return []
        return ConfigParser.SafeConfigParser.options(self, section)

    def read(self, paths, recurse=True):
        r"""
        .. highlight:: ini

        Overrides the ``RawConfigParser`` method by implementing the following
        logic:

        Prior to actually reading the contents of the file(s) specified by
        ``paths`` into the current config instance, a recursive algorithm will
        inspect the config found in the file(s) to see if additional config
        file(s) are to be included.  All config files, whether specified
        directly by ``paths`` or indirectly by way of primary configuration,
        are finally read into the current config instance in the proper order
        so that cascading works as expected.

        If you pass ``recurse=False`` to this method then none of the magical
        inclusion logic will happen at all.

        Note that when a config file indicates that another file(s) is to be
        included, the referenced file will be read into this config instance
        *before* the original (primary) file is read into it.  A convenient
        setup then could be to maintain a "site-wide" config file, shared on
        the network, including something like this::

           # //file-server/share/edbob/site.conf
           #
           # This file contains settings relevant to all machines on the
           # network.  Mail and logging configuration at least would be good
           # candidates for inclusion here.

           [edbob.mail]
           smtp.server = mail.example.com
           # smtp.username = user
           # smtp.password = pass
           sender.default = noreply@example.com
           recipients.default = ['tech-support@example.com']

           [loggers]
           keys = root, edbob

           # ...etc.  The bulk of logging configuration would go here.

        Then a config file local to a particular machine could look something
        like this::

           # C:\ProgramData\edbob\edbob.conf
           #
           # This file contains settings specific to the local machine.

           [edbob]
           include_config = [r'\\file-server\share\edbob\site.conf']

           # Add any local app config here, e.g. connection details for a
           # database, etc.

           # All logging config is inherited from the site-wide file, except
           # we'll override the output file location so that it remains local.
           # And maybe we need the level bumped up while we troubleshoot
           # something.

           [handler_file]
           args = (r'C:\ProgramData\edbob\edbob.log', 'a')

           [logger_edbob]
           level = DEBUG

        There is no right way to do this of course; the need should drive the
        method.  Since recursion is used, there is also no real limit to how
        you go about it.  A config file specific to a particular app on a
        particular machine can further include a config file specific to the
        local user on that machine, which in turn can include a file specific
        to the local machine generally, which could then include one or more
        site-wide files, etc.  Or the "most specific" (initially read; primary)
        config file could indicate which other files to include for every level
        of that, in which case recursion would be less necessary (though still
        technically used).
        """

        if isinstance(paths, basestring):
            paths = [paths]
        for path in paths:
            self.read_path(path, recurse=recurse)
        return self.paths_loaded

    def read_path(self, path, recurse=True):
        """
        .. highlight:: ini
        
        Reads a "single" config file into the instance.  If ``recurse`` is ``True``,
        *and* the config file references other "parent" config file(s), then the
        parent(s) are read also in recursive fashion.

        "Parent" config file paths may be specified in this way::
        
           [edbob]
           include_config = [
                r'\\file-server\share\edbob\site.conf',
                r'C:\ProgramData\edbob\special-stuff.conf',
                ]

        See :meth:`read()` for more information.
        """

        if path in self.paths_attempted:
            return

        self.paths_attempted.append(path)
        log.debug("Reading config file: %s" % path)
        if not os.path.exists(path):
            log.debug("File doesn't exist")
            return
        config = ConfigParser.SafeConfigParser(dict(
                here=os.path.abspath(os.path.dirname(path))))
        if not config.read(path):
            log.debug("Read failed")
            return
        include = None
        if recurse:
            if (config.has_section('edbob') and
                config.has_option('edbob', 'include_config')):
                include = config.get('edbob', 'include_config')
                if include:
                    for p in eval(include):
                        self.read_path(os.path.abspath(p))
        ConfigParser.SafeConfigParser.read(self, path)
        if include:
            self.remove_option('edbob', 'include_config')
        self.paths_loaded.append(path)
        log.info("Read config file: %s" % path)

    def read_service(self, service, paths):
        """
        "Special" version of :meth:`read()` which will first inspect the
        file(s) for a service-specific directive, the presence of which
        indicates the *true* config file to be used for the service.

        This method is pretty much a hack to get around certain limitations of
        Windows service implementations; it is not used otherwise.
        """

        config = ConfigParser.SafeConfigParser()
        config.read(paths)
        
        if (config.has_section('edbob.service_config')
            and config.has_option('edbob.service_config', service)):
            paths = eval(config.get('edbob.service_config', service))

        self.read(paths, recurse=True)

    def require(self, section, option, msg=None):
        """
        Convenience method which will raise an exception if the given option
        does not exist.  ``msg`` can be used to override (some of) the error
        text.
        """

        value = self.get(section, option)
        if value:
            return value
        raise exceptions.ConfigError(section, option, msg)

    def save(self, filename, create_dir=True):
        """
        Saves the current config contents to a file.  Optionally can create the
        parent folder(s) as necessary.
        """

        config_folder = os.path.dirname(filename)
        if create_dir and not os.path.exists(config_folder):
            os.makedirs(config_folder)
        config_file = open(filename, 'w')
        self.write(config_file)
        config_file.close()

    def set(self, section, option, value):
        """
        Overrides ``ConfigParser.SafeConfigParser.set()`` so that ``section``
        is created if it doesn't already exist, instead of raising an error.
        """

        if not self.has_section(section):
            self.add_section(section)
        ConfigParser.SafeConfigParser.set(self, section, option, value)


def default_system_paths(appname):
    r"""
    Returns a list of default system-level config file paths for the given
    ``appname``, according to ``sys.platform``.

    For example, assuming an app name of ``'rattail'``, the following would be
    returned:

    ``win32``:
       * ``<COMMON_APPDATA>\rattail.conf``
       * ``<COMMON_APPDATA>\rattail\rattail.conf``

    Any other platform:
       * ``/etc/rattail.conf``
       * ``/etc/rattail/rattail.conf``
       * ``/usr/local/etc/rattail.conf``
       * ``/usr/local/etc/rattail/rattail.conf``
    """

    if sys.platform == 'win32':
        from win32com.shell import shell, shellcon
        return [
            os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_COMMON_APPDATA), '%s.conf' % appname),
            os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_COMMON_APPDATA), appname, '%s.conf' % appname),
            ]

    return [
        '/etc/%s.conf' % appname,
        '/etc/%s/%s.conf' % (appname, appname),
        '/usr/local/etc/%s.conf' % appname,
        '/usr/local/etc/%s/%s.conf' % (appname, appname),
        ]


def default_user_paths(appname):
    r"""
    Returns a list of default user-level config file paths for the given
    ``appname``, according to ``sys.platform``.

    For example, assuming an app name of ``'rattail'``, the following would be
    returned:

    ``win32``:
       * ``<APPDATA>\rattail.conf``
       * ``<APPDATA>\rattail\rattail.conf``

    Any other platform:
       * ``~/.rattail.conf``
       * ``~/.rattail/rattail.conf``
    """

    if sys.platform == 'win32':
        from win32com.shell import shell, shellcon
        return [
            os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_APPDATA), '%s.conf' % appname),
            os.path.join(shell.SHGetSpecialFolderPath(
                    0, shellcon.CSIDL_APPDATA), appname, '%s.conf' % appname),
            ]

    return [
        os.path.expanduser('~/.%s.conf' % appname),
        os.path.expanduser('~/.%s/%s.conf' % (appname, appname)),
        ]



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.subscribers`` -- Subscribers
"""

from pyramid import threadlocal
from pyramid.security import authenticated_userid

import edbob
from edbob.db.auth import has_permission
from edbob.pyramid import helpers
from edbob.pyramid import Session


def before_render(event):
    """
    Adds goodies to the global template renderer context.
    """

    request = event.get('request') or threadlocal.get_current_request()

    renderer_globals = event
    renderer_globals['h'] = helpers
    renderer_globals['url'] = request.route_url
    renderer_globals['edbob'] = edbob
    renderer_globals['Session'] = Session


def context_found(event):
    """
    This hook attaches various attributes and methods to the ``request``
    object.  Specifically:

    The :class:`edbob.User` instance currently logged-in (if indeed there is
    one) is attached as ``request.user``.

    A ``request.has_perm()`` method is attached, which is a shortcut for
    :func:`edbob.db.auth.has_permission()`.

    A ``request.get_referrer()`` method is attached, which contains some
    convenient logic for determining the referring URL.

    The ``request.get_setting()`` and ``request.save_setting()`` methods are
    attached, which are shortcuts for :func:`edbob.get_setting()` and
    :func:`edbob.save_setting()`, respectively.
    """

    request = event.request

    request.user = None
    uuid = authenticated_userid(request)
    if uuid:
        request.user = Session.query(edbob.User).get(uuid)

    def has_perm(perm):
        return has_permission(request.user, perm, session=Session())
    request.has_perm = has_perm

    def has_any_perm(perms):
        for perm in perms:
            if has_permission(request.user, perm, session=Session()):
                return True
        return False
    request.has_any_perm = has_any_perm

    def get_referrer(default=None):
        if request.params.get('referrer'):
            return request.params['referrer']
        if request.session.get('referrer'):
            return request.session.pop('referrer')
        referrer = request.referrer
        if not referrer or referrer == request.current_route_url():
            if default:
                referrer = default
            else:
                referrer = request.route_url('home')
        return referrer
    request.get_referrer = get_referrer

    def get_setting(name):
        return edbob.get_setting(name, Session())
    request.get_setting = get_setting

    def save_setting(name, value):
        edbob.save_setting(name, value, Session())
    request.save_setting = save_setting


def includeme(config):
    config.add_subscriber('edbob.pyramid.subscribers:before_render',
                          'pyramid.events.BeforeRender')
    config.add_subscriber('edbob.pyramid.subscribers.context_found',
                          'pyramid.events.ContextFound')



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

import wx

# begin wxGlade: extracode
# end wxGlade

import edbob


class GenerateUuidDialog(wx.Dialog):
    def __init__(self, *args, **kwds):
        # begin wxGlade: GenerateUuidDialog.__init__
        kwds["style"] = wx.DEFAULT_DIALOG_STYLE
        wx.Dialog.__init__(self, *args, **kwds)
        self.label_1 = wx.StaticText(self, -1, "&UUID:")
        self.Uuid = wx.TextCtrl(self, -1, "", style=wx.TE_READONLY)
        self.Generate = wx.Button(self, -1, "&Generate UUID")
        self.Close = wx.Button(self, wx.ID_OK, "&Close")

        self.__set_properties()
        self.__do_layout()

        self.Bind(wx.EVT_BUTTON, self.OnGenerateButton, self.Generate)
        # end wxGlade

        self.GenerateUuid()

    def __set_properties(self):
        # begin wxGlade: GenerateUuidDialog.__set_properties
        self.SetTitle("Generate UUID")
        self.Uuid.SetMinSize((300, -1))
        # end wxGlade

    def __do_layout(self):
        # begin wxGlade: GenerateUuidDialog.__do_layout
        sizer_1 = wx.BoxSizer(wx.VERTICAL)
        sizer_2 = wx.BoxSizer(wx.VERTICAL)
        sizer_3 = wx.BoxSizer(wx.HORIZONTAL)
        sizer_2.Add(self.label_1, 0, 0, 0)
        sizer_2.Add(self.Uuid, 0, wx.TOP|wx.EXPAND, 5)
        sizer_3.Add(self.Generate, 0, 0, 0)
        sizer_3.Add(self.Close, 0, wx.LEFT, 10)
        sizer_2.Add(sizer_3, 0, wx.TOP|wx.ALIGN_CENTER_HORIZONTAL, 10)
        sizer_1.Add(sizer_2, 1, wx.ALL|wx.EXPAND, 10)
        self.SetSizer(sizer_1)
        sizer_1.Fit(self)
        self.Layout()
        self.Centre()
        # end wxGlade

    def OnGenerateButton(self, event): # wxGlade: GenerateUuidDialog.<event_handler>
        self.GenerateUuid()
        event.Skip()

    def GenerateUuid(self):
        self.Uuid.SetValue(edbob.get_uuid())
        self.Uuid.SetSelection(-1, -1)
        self.Uuid.SetFocus()

# end of class GenerateUuidDialog


def main():
    app = wx.PySimpleApp()
    dlg = GenerateUuidDialog(None)
    dlg.ShowModal()
    dlg.Destroy()
    app.MainLoop()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.grids.search`` -- Grid Search Filters
"""

from sqlalchemy import or_

from webhelpers.html import tags
from webhelpers.html import literal

from pyramid.renderers import render
from pyramid_simpleform import Form
from pyramid_simpleform.renderers import FormRenderer

import edbob
from edbob.util import prettify


class SearchFilter(edbob.Object):
    """
    Base class and default implementation for search filters.
    """

    def __init__(self, name, label=None, **kwargs):
        edbob.Object.__init__(self, **kwargs)
        self.name = name
        self.label = label or prettify(name)

    def types_select(self):
        types = [
            ('is', "is"),
            ('nt', "is not"),
            ('lk', "contains"),
            ('nl', "doesn't contain"),
            ]
        options = []
        filter_map = self.search.filter_map[self.name]
        for value, label in types:
            if value in filter_map:
                options.append((value, label))
        return tags.select('filter_type_'+self.name,
                           self.search.config.get('filter_type_'+self.name),
                           options, class_='filter-type')

    def value_control(self):
        return tags.text(self.name, self.search.config.get(self.name))


class BooleanSearchFilter(SearchFilter):
    """
    Boolean search filter.
    """

    def value_control(self):
        return tags.select(self.name, self.search.config.get(self.name),
                           ["True", "False"])


class SearchForm(Form):
    """
    Generic form class which aggregates :class:`SearchFilter` instances.
    """

    def __init__(self, request, filter_map, config, *args, **kwargs):
        super(SearchForm, self).__init__(request, *args, **kwargs)
        self.filter_map = filter_map
        self.config = config
        self.filters = {}

    def add_filter(self, filter_):
        filter_.search = self
        self.filters[filter_.name] = filter_


class SearchFormRenderer(FormRenderer):
    """
    Renderer for :class:`SearchForm` instances.
    """

    def __init__(self, form, *args, **kwargs):
        super(SearchFormRenderer, self).__init__(form, *args, **kwargs)
        self.request = form.request
        self.filters = form.filters
        self.config = form.config

    def add_filter(self, visible):
        options = ['add a filter']
        for f in self.sorted_filters():
            options.append((f.name, f.label))
        return self.select('add-filter', options,
                           style='display: none;' if len(visible) == len(self.filters) else None)

    def checkbox(self, name, checked=None, **kwargs):
        if name.startswith('include_filter_'):
            if checked is None:
                checked = self.config[name]
            return tags.checkbox(name, checked=checked, **kwargs)
        if checked is None:
            checked = False
        return super(SearchFormRenderer, self).checkbox(name, checked=checked, **kwargs)

    def render(self, **kwargs):
        kwargs['search'] = self
        return literal(render('/grids/search.mako', kwargs))

    def sorted_filters(self):
        return sorted(self.filters.values(), key=lambda x: x.label)

    def text(self, name, **kwargs):
        return tags.text(name, value=self.config.get(name), **kwargs)


def filter_exact(field):
    """
    Convenience function which returns a filter map entry, with typical logic
    built in for "exact match" queries applied to ``field``.
    """

    return {
        'is':
            lambda q, v: q.filter(field == v) if v else q,
        'nt':
            lambda q, v: q.filter(field != v) if v else q,
        }


def filter_ilike(field):
    """
    Convenience function which returns a filter map entry, with typical logic
    built in for "ILIKE" queries applied to ``field``.
    """

    def ilike(query, value):
        if value:
            query = query.filter(field.ilike('%%%s%%' % value))
        return query

    def not_ilike(query, value):
        if value:
            query = query.filter(or_(
                    field == None,
                    ~field.ilike('%%%s%%' % value),
                    ))
        return query
    
    return {'lk': ilike, 'nl': not_ilike}


def get_filter_config(prefix, request, filter_map, **kwargs):
    """
    Returns a configuration dictionary for a search form.
    """

    config = {}

    def update_config(dict_, prefix='', exclude_by_default=False):
        """
        Updates the ``config`` dictionary based on the contents of ``dict_``.
        """

        for field in filter_map:
            if prefix+'include_filter_'+field in dict_:
                include = dict_[prefix+'include_filter_'+field]
                include = bool(include) and include != '0'
                config['include_filter_'+field] = include
            elif exclude_by_default:
                config['include_filter_'+field] = False
            if prefix+'filter_type_'+field in dict_:
                config['filter_type_'+field] = dict_[prefix+'filter_type_'+field]
            if prefix+field in dict_:
                config[field] = dict_[prefix+field]

    # Update config to exclude all filters by default.
    for field in filter_map:
        config['include_filter_'+field] = False

    # Update config to honor default settings.
    config.update(kwargs)

    # Update config with data cached in session.
    update_config(request.session, prefix=prefix+'.')

    # Update config with data from GET/POST request.
    if request.params.get('filters') == 'true':
        update_config(request.params, exclude_by_default=True)

    # Cache filter data in session.
    for key in config:
        if (not key.startswith('filter_factory_')
            and not key.startswith('filter_label_')):
            request.session[prefix+'.'+key] = config[key]

    return config


def get_filter_map(cls, exact=[], ilike=[], **kwargs):
    """
    Convenience function which returns a "filter map" for ``cls``.

    ``exact``, if provided, should be a list of field names for which "exact"
    filtering is to be allowed.

    ``ilike``, if provided, should be a list of field names for which "ILIKE"
    filtering is to be allowed.

    Any remaining ``kwargs`` are assumed to be filter map entries themselves,
    and are added directly to the map.
    """

    fmap = {}
    for name in exact:
        fmap[name] = filter_exact(getattr(cls, name))
    for name in ilike:
        fmap[name] = filter_ilike(getattr(cls, name))
    fmap.update(kwargs)
    return fmap


def get_search_form(request, filter_map, config):
    """
    Returns a :class:`SearchForm` instance with a :class:`SearchFilter` for
    each filter in ``filter_map``, using configuration from ``config``.
    """

    search = SearchForm(request, filter_map, config)
    for field in filter_map:
        factory = config.get('filter_factory_%s' % field, SearchFilter)
        label = config.get('filter_label_%s' % field)
        search.add_filter(factory(field, label=label))
    return search


def filter_query(query, config, filter_map, join_map):
    """
    Filters ``query`` according to ``config`` and ``filter_map``.  ``join_map``
    is used, if necessary, to join additional tables to the base query.  The
    filtered query is returned.
    """

    joins = config.setdefault('joins', [])
    for key in config:
        if key.startswith('include_filter_') and config[key]:
            field = key[15:]
            if field in join_map and field not in joins:
                query = join_map[field](query)
                joins.append(field)
            value = config.get(field)
            if value:
                fmap = filter_map[field]
                filt = fmap[config['filter_type_'+field]]
                query = filt(query, value)
    return query



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.autocomplete`` -- Autocomplete View
"""

from edbob.pyramid import Session
from edbob.pyramid.views.core import View
from edbob.util import requires_impl


__all__ = ['AutocompleteView']


class AutocompleteView(View):

    @property
    @requires_impl(is_property=True)
    def mapped_class(self):
        raise NotImplementedError

    @property
    @requires_impl(is_property=True)
    def fieldname(self):
        raise NotImplementedError

    def filter_query(self, q):
        return q

    def make_query(self, query):
        q = Session.query(self.mapped_class)
        q = self.filter_query(q)
        q = q.filter(getattr(self.mapped_class, self.fieldname).ilike('%%%s%%' % query))
        q = q.order_by(getattr(self.mapped_class, self.fieldname))
        return q

    def query(self, query):
        return self.make_query(query)

    def display(self, instance):
        return getattr(instance, self.fieldname)

    def __call__(self):
        query = self.request.params['query']
        objs = self.query(query).all()
        data = dict(
            query=query,
            suggestions=[self.display(x) for x in objs],
            data=[x.uuid for x in objs],
            )
        return data



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.roles`` -- Role Views
"""

from pyramid.httpexceptions import HTTPFound

import formalchemy
from webhelpers.html import tags
from webhelpers.html.builder import HTML

import edbob
from edbob.db import auth
from edbob.pyramid import Session
from edbob.pyramid.views import SearchableAlchemyGridView, CrudView


default_permissions = [

    ("People", [
            ('people.list',         "List People"),
            ('people.read',         "View Person"),
            ('people.create',       "Create Person"),
            ('people.update',       "Edit Person"),
            ('people.delete',       "Delete Person"),
            ]),

    ("Roles", [
            ('roles.list',          "List Roles"),
            ('roles.read',          "View Role"),
            ('roles.create',        "Create Role"),
            ('roles.update',        "Edit Role"),
            ('roles.delete',        "Delete Role"),
            ]),

    ("Users", [
            ('users.list',          "List Users"),
            ('users.read',          "View User"),
            ('users.create',        "Create User"),
            ('users.update',        "Edit User"),
            ('users.delete',        "Delete User"),
            ]),
    ]


class RolesGrid(SearchableAlchemyGridView):

    mapped_class = edbob.Role
    config_prefix = 'roles'
    sort = 'name'

    def filter_map(self):
        return self.make_filter_map(ilike=['name'])

    def filter_config(self):
        return self.make_filter_config(
            include_filter_name=True,
            filter_type_name='lk')

    def sort_map(self):
        return self.make_sort_map('name')

    def grid(self):
        g = self.make_grid()
        g.configure(
            include=[
                g.name,
                ],
            readonly=True)
        if self.request.has_perm('roles.read'):
            g.clickable = True
            g.click_route_name = 'role.read'
        if self.request.has_perm('roles.update'):
            g.editable = True
            g.edit_route_name = 'role.update'
        if self.request.has_perm('roles.delete'):
            g.deletable = True
            g.delete_route_name = 'role.delete'
        return g


class PermissionsField(formalchemy.Field):

    def sync(self):
        if not self.is_readonly():
            role = self.model
            role.permissions = self.renderer.deserialize()


def PermissionsFieldRenderer(permissions, *args, **kwargs):

    perms = permissions
    
    class PermissionsFieldRenderer(formalchemy.FieldRenderer):

        permissions = perms

        def deserialize(self):
            perms = []
            i = len(self.name) + 1
            for key in self.params:
                if key.startswith(self.name):
                    perms.append(key[i:])
            return perms

        def _render(self, readonly=False, **kwargs):
            role = self.field.model
            admin = auth.administrator_role(Session())
            if role is admin:
                html = HTML.tag('p', c="This is the administrative role; "
                                "it has full access to the entire system.")
                if not readonly:
                    html += tags.hidden(self.name, value='') # ugly hack..or good idea?
            else:
                html = ''
                for group, perms in self.permissions:
                    inner = HTML.tag('p', c=group)
                    for perm, title in perms:
                        checked = auth.has_permission(
                            role, perm, include_guest=False, session=Session())
                        if readonly:
                            span = HTML.tag('span', c="[X]" if checked else "[ ]")
                            inner += HTML.tag('p', class_='perm', c=span + ' ' + title)
                        else:
                            inner += tags.checkbox(self.name + '-' + perm,
                                                   checked=checked, label=title)
                    html += HTML.tag('div', class_='group', c=inner)
            return html

        def render(self, **kwargs):
            return self._render(**kwargs)

        def render_readonly(self, **kwargs):
            return self._render(readonly=True, **kwargs)

    return PermissionsFieldRenderer


class RoleCrud(CrudView):

    mapped_class = edbob.Role
    home_route = 'roles'
    permissions = default_permissions

    def fieldset(self, role):
        fs = self.make_fieldset(role)
        fs.append(PermissionsField(
                'permissions',
                renderer=PermissionsFieldRenderer(self.permissions)))
        fs.configure(
            include=[
                fs.name,
                fs.permissions,
                ])
        return fs

    def pre_delete(self, model):
        admin = auth.administrator_role(Session())
        guest = auth.guest_role(Session())
        if model in (admin, guest):
            self.request.session.flash("You may not delete the %s role." % str(model), 'error')
            return HTTPFound(location=self.request.get_referrer())


def includeme(config):
    
    config.add_route('roles', '/roles')
    config.add_view(RolesGrid, route_name='roles',
                    renderer='/roles/index.mako',
                    permission='roles.list')

    settings = config.get_settings()
    perms = settings.get('edbob.permissions')
    if perms:
        RoleCrud.permissions = perms

    config.add_route('role.create', '/roles/new')
    config.add_view(RoleCrud, attr='create', route_name='role.create',
                    renderer='/roles/crud.mako',
                    permission='roles.create')

    config.add_route('role.read', '/roles/{uuid}')
    config.add_view(RoleCrud, attr='read', route_name='role.read',
                    renderer='/roles/crud.mako',
                    permission='roles.read')

    config.add_route('role.update', '/roles/{uuid}/edit')
    config.add_view(RoleCrud, attr='update', route_name='role.update',
                    renderer='/roles/crud.mako',
                    permission='roles.update')

    config.add_route('role.delete', '/roles/{uuid}/delete')
    config.add_view(RoleCrud, attr='delete', route_name='role.delete',
                    permission='roles.delete')



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.form`` -- Generic Form View
"""

import transaction
from pyramid.httpexceptions import HTTPFound

from edbob.pyramid import Session
from edbob.pyramid.forms import SimpleForm
from edbob.util import requires_impl


__all__ = ['FormView']


class FormView(object):
    """
    This view provides basic form processing goodies.
    """

    route = None
    url = None
    template = None
    permission = None

    def __init__(self, request):
        self.request = request

    def __call__(self):
        """
        Callable for the view.  This method creates the underlying form and
        processes data if any was submitted.
        """

        f = self.form(self.request)
        if not f.readonly and self.request.POST:
            f.rebind(data=self.request.params)
            if f.validate():

                with transaction.manager:
                    f.save(Session)
                    Session.flush()
                    self.request.session.flash('The book "%s" has been loaned.' % f.book.value)

                if self.request.params.get('keep-going') == '1':
                    return HTTPFound(location=self.request.current_route_url())

                return HTTPFound(location=f.home_url)

        return {'form': f}

    def make_form(self, request, **kwargs):
        """
        Returns a :class:`edbob.pyramid.forms.Form` instance.
        """
        template = kwargs.pop('template', self.template or '%s.mako' % self.url)
        return SimpleForm(request, template=template, **kwargs)

    def form(self, request):
        """
        Should create and return a :class:`edbob.pyramid.forms.Form` instance
        for the view.
        """
        return self.make_form(request)

    @classmethod
    def add_route(cls, config, **kwargs):
        route = kwargs.get('route', cls.route)
        url = kwargs.get('url', cls.url)
        permission = kwargs.get('permission', cls.permission or route)
        template = kwargs.get('template', cls.template or '%s.mako' % url)
        config.add_route(route, url)
        config.add_view(cls, route_name=route, renderer=template,
                        permission=permission, http_cache=0)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.time`` -- Date & Time Utilities
"""

import datetime
import pytz
import logging
import warnings

import edbob


__all__ = ['utc_time', 'local_time']

timezones = {}

log = logging.getLogger(__name__)


def init(config):
    """
    Reads configuration to become aware of all timezones which may concern the
    application.

    .. highlight:: ini

    The bare minimum configuration required is the ``zone.local`` setting::

       [edbob.time]
       zone.local = US/Pacific

    Multiple timezones may be configured like so::

       [edbob.time]
       zone.local = America/Los_Angeles
       zone.head_office = America/New_York
       zone.that_other_place = Asia/Manila
       zone.some_app = US/Central

    See `Wikipedia
    <http://en.wikipedia.org/wiki/List_of_tz_database_time_zones>`_ for a
    (presumably) full list of valid timezone names.

    .. note::
       A ``zone.utc`` option is automatically created for you; there is no need
       to define it.
    """

    for key in config.options('edbob.time'):
        if key.startswith('zone.'):
            tz = config.get('edbob.time', key)
            if tz:
                key = key[5:]
                log.info("'%s' timezone set to '%s'" % (key, tz))
                set_timezone(tz, key)

    if 'local' not in timezones:
        tz = config.get('edbob.time', 'timezone')
        if tz:
            warnings.warn("Config option 'timezone' in 'edbob.time' section is deprecated.  "
                          "Please set 'zone.local' instead.",
                          DeprecationWarning)
            set_timezone(tz)
        else:
            log.warning("'local' timezone not configured; falling back to 'America/Chicago'")
            set_timezone('America/Chicago')

    set_timezone('UTC', 'utc')


def get_timezone(key='local'):
    """
    Returns the timezone referenced by ``key``.
    """

    if key not in timezones:
        edbob.config.require('edbob.time', 'zone.%s' % key)
    return timezones[key]


def set_timezone(tz, key='local'):
    """
    Stores a timezone in the global dictionary, using ``key``.

    ``tz`` must be a valid "Olson" time zone name, e.g. ``'US/Central'``.  See
    `Wikipedia <http://en.wikipedia.org/wiki/List_of_tz_database_time_zones>`_
    for a (presumably) full list of valid names.
    """

    timezones[key] = pytz.timezone(tz)


def local_time(stamp=None, from_='local', naive=False):
    """
    Returns a :class:`datetime.datetime` instance, with its ``tzinfo`` member
    set to the timezone referenced by the key ``'local'``.  If ``naive`` is
    ``True``, the result is stripped of its ``tzinfo`` member.

    If ``stamp`` is provided, and it is not already "aware," then its value is
    interpreted as being local to the timezone referenced by ``from_``.

    If ``stamp`` is not provided, the current time is assumed.
    """

    if not stamp:
        stamp = utc_time()
    elif not stamp.tzinfo:
        stamp = localize(stamp, from_=from_)
    return localize(stamp, naive=naive)


def localize(stamp, from_='local', to='local', naive=False):
    """
    Creates a "localized" version of ``stamp`` and returns it.

    ``stamp`` must be a :class:`datetime.datetime` instance.  If it is naive,
    its value is interpreted as being local to the timezone referenced by
    ``from_``.  If it is already aware (i.e. not naive), then ``from_`` is
    ignored.

    The timezone referenced by ``to`` is used to determine the final, "local"
    value for the timestamp.  If ``naive`` is ``True``, the timestamp is
    stripped of its ``tzinfo`` member before being returned.  Otherwise, it
    will remain aware of its timezone.
    """

    if not stamp.tzinfo:
        tz = get_timezone(from_)
        stamp = tz.localize(stamp)
    tz = get_timezone(to)
    stamp = stamp.astimezone(tz)
    if naive:
        stamp = stamp.replace(tzinfo=None)
    return stamp


def utc_time(stamp=None, from_='local', naive=False):
    """
    Returns a :class:`datetime.datetime` instance, with its ``tzinfo`` member
    set to the UTC timezone.  If ``naive`` is ``True``, the result is stripped
    of its ``tzinfo`` member.

    If ``stamp`` is provided, and it is not already "aware," then its value is
    interpreted as being local to the timezone referenced by ``from_``.

    If ``stamp`` is not provided, the current time is assumed.
    """

    if not stamp:
        stamp = datetime.datetime.utcnow()
        stamp = pytz.utc.localize(stamp)
    elif not stamp.tzinfo:
        stamp = localize(stamp, from_=from_)
    return localize(stamp, to='utc', naive=naive)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.filemon.linux`` -- File Monitor for Linux
"""

import sys
import os
import os.path
import signal
import logging
import pyinotify

import edbob
from edbob.filemon import get_monitor_profiles


log = logging.getLogger(__name__)


class EventHandler(pyinotify.ProcessEvent):
    """
    Event processor for file monitor daemon.
    """

    def my_init(self, actions=[], locks=False, **kwargs):
        self.actions = actions
        self.locks = locks

    def process_IN_ACCESS(self, event):
        log.debug("EventHandler: IN_ACCESS: %s" % event.pathname)

    def process_IN_ATTRIB(self, event):
        log.debug("EventHandler: IN_ATTRIB: %s" % event.pathname)

    def process_IN_CLOSE_WRITE(self, event):
        log.debug("EventHandler: IN_CLOSE_WRITE: %s" % event.pathname)
        if not self.locks:
            self.perform_actions(event.pathname)

    def process_IN_CREATE(self, event):
        log.debug("EventHandler: IN_CREATE: %s" % event.pathname)

    def process_IN_DELETE(self, event):
        log.debug("EventHandler: IN_DELETE: %s" % event.pathname)
        if self.locks and event.pathname.endswith('.lock'):
            self.perform_actions(event.pathname[:-5])

    def process_IN_MODIFY(self, event):
        log.debug("EventHandler: IN_MODIFY: %s" % event.pathname)

    def process_IN_MOVED_TO(self, event):
        log.debug("EventHandler: IN_MOVED_TO: %s" % event.pathname)
        if not self.locks:
            self.perform_actions(event.pathname)

    def perform_actions(self, path):
        for spec, func, args in self.actions:
            func(path, *args)


def get_pid_path():
    """
    Returns the path to the PID file for the file monitor daemon.
    """

    basename = os.path.basename(sys.argv[0])
    pid_path = edbob.config.get('%s.filemon' % basename, 'pid_path')
    if not pid_path:
        pid_path = '/tmp/%s_filemon.pid' % basename
    return pid_path


def start_daemon(appname, daemonize=True):
    """
    Starts the file monitor daemon.
    """

    pid_path = get_pid_path()
    if os.path.exists(pid_path):
        print "File monitor is already running"
        return

    wm = pyinotify.WatchManager()
    notifier = pyinotify.Notifier(wm)

    monitored = get_monitor_profiles(appname)

    mask = (pyinotify.IN_ACCESS | pyinotify.IN_ATTRIB
            | pyinotify.IN_CLOSE_WRITE | pyinotify.IN_CREATE
            | pyinotify.IN_DELETE | pyinotify.IN_MODIFY
            | pyinotify.IN_MOVED_TO)
    for profile in monitored.itervalues():
        for path in profile.dirs:
            wm.add_watch(path, mask, proc_fun=EventHandler(
                    actions=profile.actions, locks=profile.locks))

    if not daemonize:
        sys.stderr.write("Starting file monitor.  (Press Ctrl+C to quit.)\n")
    notifier.loop(daemonize=daemonize, pid_file=pid_path)


def stop_daemon():
    """
    Stops the file monitor daemon.
    """

    pid_path = get_pid_path()
    if not os.path.exists(pid_path):
        print "File monitor is not running"
        return

    f = open(pid_path)
    pid = f.read().strip()
    f.close()
    if not pid.isdigit():
        log.warning("stop_daemon: Found bogus PID (%s) in file: %s" % (pid, pid_path))
        return

    os.kill(int(pid), signal.SIGKILL)
    os.remove(pid_path)



#!/usr/bin/env python
# -*- coding: utf-8  -*-
################################################################################
#
#  edbob -- Pythonic Software Framework
#  Copyright © 2010-2012 Lance Edgar
#
#  This file is part of edbob.
#
#  edbob is free software: you can redistribute it and/or modify it under the
#  terms of the GNU Affero General Public License as published by the Free
#  Software Foundation, either version 3 of the License, or (at your option)
#  any later version.
#
#  edbob is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for
#  more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with edbob.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################

"""
``edbob.pyramid.views.crud`` -- CRUD View Function
"""

from pyramid.httpexceptions import HTTPFound

import formalchemy

from edbob.pyramid import Session
from edbob.pyramid.forms.formalchemy import AlchemyForm
from edbob.pyramid.views.core import View
from edbob.util import requires_impl, prettify


__all__ = ['CrudView']


class CrudView(View):

    readonly = False
    allow_successive_creates = False
    update_cancel_route = None

    @property
    @requires_impl(is_property=True)
    def mapped_class(self):
        pass

    @property
    def pretty_name(self):
        return self.mapped_class.__name__

    @property
    @requires_impl(is_property=True)
    def home_route(self):
        pass

    @property
    def home_url(self):
        return self.request.route_url(self.home_route)

    @property
    def cancel_route(self):
        return self.home_route

    @property
    def cancel_url(self):
        return self.request.route_url(self.cancel_route)

    def make_fieldset(self, model, **kwargs):
        kwargs.setdefault('session', Session())
        fieldset = formalchemy.FieldSet(model, **kwargs)
        fieldset.prettify = prettify
        return fieldset

    def fieldset(self, model):
        return self.make_fieldset(model)

    def make_form(self, model, **kwargs):
        self.creating = model is self.mapped_class
        self.updating = not self.creating

        fieldset = self.fieldset(model)
        kwargs.setdefault('pretty_name', self.pretty_name)
        kwargs.setdefault('action_url', self.request.current_route_url())
        if self.updating and self.update_cancel_route:
            kwargs.setdefault('cancel_url', self.request.route_url(
                    self.update_cancel_route, uuid=model.uuid))
        else:
            kwargs.setdefault('cancel_url', self.cancel_url)
        kwargs.setdefault('creating', self.creating)
        kwargs.setdefault('updating', self.updating)
        form = AlchemyForm(self.request, fieldset, **kwargs)

        if form.creating:
            if hasattr(self, 'create_label'):
                form.create_label = self.create_label
            if self.allow_successive_creates:
                form.allow_successive_creates = True
                if hasattr(self, 'successive_create_label'):
                    form.successive_create_label = self.successive_create_label

        return form

    def form(self, model):
        return self.make_form(model)

    def crud(self, model, readonly=False):

        if readonly:
            self.readonly = True

        form = self.form(model)
        if readonly:
            form.readonly = True

        if not form.readonly and self.request.POST:
            if form.validate():
                form.save()

                result = self.post_save(form)
                if result:
                    return result

                if form.creating:
                    self.flash_create(form.fieldset.model)
                else:
                    self.flash_update(form.fieldset.model)

                if (form.creating and form.allow_successive_creates
                    and self.request.params.get('create_and_continue')):
                    return HTTPFound(location=self.request.current_route_url())

                return HTTPFound(location=self.post_save_url(form))

            self.validation_failed(form)

        kwargs = self.template_kwargs(form)
        kwargs['form'] = form
        return kwargs

    def template_kwargs(self, form):
        return {}

    def post_save(self, form):
        pass

    def post_save_url(self, form):
        return self.home_url

    def validation_failed(self, form):
        pass

    def flash_create(self, model):
        self.request.session.flash("%s \"%s\" has been created." %
                                   (self.pretty_name, model))

    def flash_delete(self, model):
        self.request.session.flash("%s \"%s\" has been deleted." %
                                   (self.pretty_name, model))
        
    def flash_update(self, model):
        self.request.session.flash("%s \"%s\" has been updated." %
                                   (self.pretty_name, model))

    def create(self):
        return self.crud(self.mapped_class)

    def read(self):
        uuid = self.request.matchdict['uuid']
        model = Session.query(self.mapped_class).get(uuid) if uuid else None
        if not model:
            return HTTPFound(location=self.home_url)
        return self.crud(model, readonly=True)

    def update(self):
        uuid = self.request.matchdict['uuid']
        model = Session.query(self.mapped_class).get(uuid) if uuid else None
        assert model
        return self.crud(model)

    def pre_delete(self, model):
        pass

    def post_delete(self, model):
        pass

    def delete(self):
        uuid = self.request.matchdict['uuid']
        model = Session.query(self.mapped_class).get(uuid) if uuid else None
        assert model
        result = self.pre_delete(model)
        if result:
            return result
        Session.delete(model)
        Session.flush() # Don't set flash message if delete fails.
        self.post_delete(model)
        self.flash_delete(model)
        return HTTPFound(location=self.home_url)
